# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a cylinder."""

from functools import cached_property

from beartype import beartype as check_input_types
import numpy as np
from pint import Quantity

from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Z
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.measurements import Distance
from ansys.geometry.core.shapes.curves.circle import Circle
from ansys.geometry.core.shapes.curves.line import Line
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.typing import Real, RealSequence


class Cylinder(Surface):
    """Provides 3D cylinder representation.

    Parameters
    ----------
    origin : ~numpy.ndarray | RealSequence | Point3D
        Origin of the cylinder.
    radius : ~pint.Quantity | Distance | Real
        Radius of the cylinder.
    reference : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
        X-axis direction.
    axis : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
        Z-axis direction.
    """

    @check_input_types
    def __init__(
        self,
        origin: np.ndarray | RealSequence | Point3D,
        radius: Quantity | Distance | Real,
        reference: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_X,
        axis: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_Z,
    ):
        """Initialize the ``Cylinder`` class."""
        self._origin = Point3D(origin) if not isinstance(origin, Point3D) else origin
        self._reference = (
            UnitVector3D(reference) if not isinstance(reference, UnitVector3D) else reference
        )
        self._axis = UnitVector3D(axis) if not isinstance(axis, UnitVector3D) else axis
        self._axis = UnitVector3D(axis) if not isinstance(axis, UnitVector3D) else axis
        if not self._reference.is_perpendicular_to(self._axis):
            raise ValueError("Cylinder reference (dir_x) and axis (dir_z) must be perpendicular.")

        self._radius = radius if isinstance(radius, Distance) else Distance(radius)
        if self._radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

    @property
    def origin(self) -> Point3D:
        """Origin of the cylinder."""
        return self._origin

    @property
    def radius(self) -> Quantity:
        """Radius of the cylinder."""
        return self._radius.value

    @property
    def dir_x(self) -> UnitVector3D:
        """X-direction of the cylinder."""
        return self._reference

    @property
    def dir_y(self) -> UnitVector3D:
        """Y-direction of the cylinder."""
        return self.dir_z.cross(self.dir_x)

    @property
    def dir_z(self) -> UnitVector3D:
        """Z-direction of the cylinder."""
        return self._axis

    def surface_area(self, height: Quantity | Distance | Real) -> Quantity:
        """Get the surface area of the cylinder.

        Parameters
        ----------
        height : ~pint.Quantity | Distance | Real
            Height to bound the cylinder at.

        Returns
        -------
        ~pint.Quantity
            Surface area of the temporarily bounded cylinder.

        Notes
        -----
        By nature, a cylinder is infinite. If you want to get the surface area,
        you must bound it by a height. Normally a cylinder surface is not closed
        (does not have "caps" on the ends). This method assumes that the cylinder
        is closed for the purpose of getting the surface area.
        """
        height = height if isinstance(height, Distance) else Distance(height)
        if height.value <= 0:
            raise ValueError("Height must be a real positive value.")

        return 2 * np.pi * self.radius * height.value + 2 * np.pi * self.radius**2

    def volume(self, height: Quantity | Distance | Real) -> Quantity:
        """Get the volume of the cylinder.

        Parameters
        ----------
        height : ~pint.Quantity | Distance | Real
            Height to bound the cylinder at.

        Returns
        -------
        ~pint.Quantity
            Volume of the temporarily bounded cylinder.

        Notes
        -----
        By nature, a cylinder is infinite. If you want to get the surface area,
        you must bound it by a height. Normally a cylinder surface is not closed
        (does not have "caps" on the ends). This method assumes that the cylinder
        is closed for the purpose of getting the surface area.
        """
        height = height if isinstance(height, Distance) else Distance(height)
        if height.value <= 0:
            raise ValueError("Height must be a real positive value.")

        return np.pi * self.radius**2 * height.value

    def transformed_copy(self, matrix: Matrix44) -> "Cylinder":
        """Create a transformed copy of the cylinder from a transformation matrix.

        Parameters
        ----------
        matrix : Matrix44
            4X4 transformation matrix to apply to the cylinder.

        Returns
        -------
        Cylinder
            New cylinder that is the transformed copy of the original cylinder.
        """
        new_point = self.origin.transform(matrix)
        new_reference = self._reference.transform(matrix)
        new_axis = self._axis.transform(matrix)
        return Cylinder(
            new_point,
            self.radius,
            UnitVector3D(new_reference[0:3]),
            UnitVector3D(new_axis[0:3]),
        )

    def mirrored_copy(self) -> "Cylinder":
        """Create a mirrored copy of the cylinder along the y-axis.

        Returns
        -------
        Cylinder
            New cylinder that is a mirrored copy of the original cylinder.
        """
        return Cylinder(self.origin, self.radius, -self._reference, -self._axis)

    @check_input_types
    def __eq__(self, other: "Cylinder") -> bool:
        """Equals operator for the ``Cylinder`` class."""
        return (
            self._origin == other._origin
            and self._radius == other._radius
            and self._reference == other._reference
            and self._axis == other._axis
        )

    def evaluate(self, parameter: ParamUV) -> "CylinderEvaluation":
        """Evaluate the cylinder at the given parameters.

        Parameters
        ----------
        parameter : ParamUV
            Parameters (u,v) to evaluate the cylinder at.

        Returns
        -------
        CylinderEvaluation
            Resulting evaluation.
        """
        return CylinderEvaluation(self, parameter)

    def project_point(self, point: Point3D) -> "CylinderEvaluation":
        """Project a point onto the cylinder and evaluate the cylinder.

        Parameters
        ----------
        point : Point3D
            Point to project onto the cylinder.

        Returns
        -------
        CylinderEvaluation
            Resulting evaluation.
        """
        circle = Circle(self.origin, self.radius, self.dir_x, self.dir_z)
        u = circle.project_point(point).parameter

        line = Line(self.origin, self.dir_z)
        v = line.project_point(point).parameter

        return CylinderEvaluation(self, ParamUV(u, v))

    def parameterization(self) -> tuple[Parameterization, Parameterization]:
        """Parameterize the cylinder surface as a tuple (U and V respectively).

        The U parameter specifies the clockwise angle around the axis (right-hand
        corkscrew law), with a zero parameter at ``dir_x`` and a period of 2*pi.

        The V parameter specifies the distance along the axis, with a zero parameter at
        the XY plane of the cylinder.

        Returns
        -------
        tuple[Parameterization, Parameterization]
            Information about how a cylinder's u and v parameters are parameterized, respectively.
        """
        u = Parameterization(ParamForm.PERIODIC, ParamType.CIRCULAR, Interval(0, 2 * np.pi))
        v = Parameterization(ParamForm.OPEN, ParamType.LINEAR, Interval(-np.inf, np.inf))

        return (u, v)

    def contains_param(self, param_uv: ParamUV) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")


class CylinderEvaluation(SurfaceEvaluation):
    """Provides evaluation of a cylinder at given parameters.

    Parameters
    ----------
    cylinder: ~ansys.geometry.core.shapes.surfaces.cylinder.Cylinder
        Cylinder to evaluate.
    parameter: ParamUV
        Parameters (u, v) to evaluate the cylinder at.
    """

    def __init__(self, cylinder: Cylinder, parameter: ParamUV) -> None:
        """Initialize the ``CylinderEvaluation`` class."""
        self._cylinder = cylinder
        self._parameter = parameter

    @property
    def cylinder(self) -> Cylinder:
        """Cylinder being evaluated."""
        return self._cylinder

    @property
    def parameter(self) -> ParamUV:
        """Parameter that the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """Position of the evaluation.

        Returns
        -------
        Point3D
            Point that lies on the cylinder at this evaluation.
        """
        return (
            self.cylinder.origin
            + self.cylinder.radius.m * self.__cylinder_normal
            + self.parameter.v * self.cylinder.dir_z
        )

    @cached_property
    def normal(self) -> UnitVector3D:
        """Normal to the surface.

        Returns
        -------
        UnitVector3D
            Normal unit vector to the cylinder at this evaluation.
        """
        return UnitVector3D(self.__cylinder_normal)

    @cached_property
    def __cylinder_normal(self) -> Vector3D:
        """Normal to the surface.

        Returns
        -------
        UnitVector3D
            Normal unit vector to the cylinder at this evaluation.
        """
        return (
            np.cos(self.parameter.u) * self.cylinder.dir_x
            + np.sin(self.parameter.u) * self.cylinder.dir_y
        )

    @cached_property
    def __cylinder_tangent(self) -> Vector3D:
        """Private tangent helper method."""
        return (
            -np.sin(self.parameter.u) * self.cylinder.dir_x
            + np.cos(self.parameter.u) * self.cylinder.dir_y
        )

    @cached_property
    def u_derivative(self) -> Vector3D:
        """First derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the U parameter.
        """
        return self.cylinder.radius.m * self.__cylinder_tangent

    @cached_property
    def v_derivative(self) -> Vector3D:
        """First derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the V parameter.
        """
        return self.cylinder.dir_z

    @cached_property
    def uu_derivative(self) -> Vector3D:
        """Second derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U parameter.
        """
        return -self.cylinder.radius.m * self.__cylinder_normal

    @cached_property
    def uv_derivative(self) -> Vector3D:
        """Second derivative with respect to the U and V parameters.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U and v parameters.
        """
        return Vector3D([0, 0, 0])

    @cached_property
    def vv_derivative(self) -> Vector3D:
        """Second derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the V parameter.
        """
        return Vector3D([0, 0, 0])

    @cached_property
    def min_curvature(self) -> Real:
        """Minimum curvature of the cylinder.

        Returns
        -------
        Real
            Minimum curvature of the cylinder.
        """
        return 0

    @cached_property
    def min_curvature_direction(self) -> UnitVector3D:
        """Minimum curvature direction.

        Returns
        -------
        UnitVector3D
            Mminimum curvature direction.
        """
        return UnitVector3D(self.cylinder.dir_z)

    @cached_property
    def max_curvature(self) -> Real:
        """Maximum curvature of the cylinder.

        Returns
        -------
        Real
            Maximum curvature of the cylinder.
        """
        return 1.0 / self.cylinder.radius.m

    @cached_property
    def max_curvature_direction(self) -> UnitVector3D:
        """Maximum curvature direction.

        Returns
        -------
        UnitVector3D
            Maximum curvature direction.
        """
        return UnitVector3D(self.u_derivative)



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides various option classes."""

from dataclasses import asdict, dataclass

from beartype import beartype as check_input_types

from ansys.geometry.core.typing import Real


@dataclass
class ImportOptions:
    """Import options when opening a file.

    Parameters
    ----------
    cleanup_bodies : bool = False
        Simplify geometry and clean up topology.
    import_coordinate_systems : bool = False
        Import coordinate systems.
    import_curves : bool = False
        Import curves.
    import_hidden_components_and_geometry : bool = False
        Import hidden components and geometry.
    import_names : bool = False
        Import names of bodies and curves.
    import_planes : bool = False
        Import planes.
    import_points : bool = False
        Import points.
    """

    cleanup_bodies: bool = False
    import_coordinate_systems: bool = False
    import_curves: bool = False
    import_hidden_components_and_geometry: bool = False
    import_names: bool = False
    import_planes: bool = False
    import_points: bool = False

    def to_dict(self):
        """Provide the dictionary representation of the ImportOptions class."""
        return {k: bool(v) for k, v in asdict(self).items()}


class TessellationOptions:
    """Provides options for getting tessellation.

    Parameters
    ----------
    surface_deviation : Real
        The maximum deviation from the true surface position.
    angle_deviation : Real
        The maximum deviation from the true surface normal, in radians.
    max_aspect_ratio : Real, default=0.0
        The maximum aspect ratio of facets.
    max_edge_length : Real, default=0.0
        The maximum facet edge length.
    watertight : bool, default=False
        Whether triangles on opposite sides of an edge should match.
    """

    @check_input_types
    def __init__(
        self,
        surface_deviation: Real,
        angle_deviation: Real,
        max_aspect_ratio: Real = 0.0,
        max_edge_length: Real = 0.0,
        watertight: bool = False,
    ):
        """Initialize ``TessellationOptions`` class."""
        self._surface_deviation = surface_deviation
        self._angle_deviation = angle_deviation
        self._max_aspect_ratio = max_aspect_ratio
        self._max_edge_length = max_edge_length
        self._watertight = watertight

    @property
    def surface_deviation(self) -> Real:
        """Surface Deviation.

        The maximum deviation from the true surface position.
        """
        return self._surface_deviation

    @property
    def angle_deviation(self) -> Real:
        """Angle deviation.

        The maximum deviation from the true surface normal, in radians.
        """
        return self._angle_deviation

    @property
    def max_aspect_ratio(self) -> Real:
        """Maximum aspect ratio.

        The maximum aspect ratio of facets.
        """
        return self._max_aspect_ratio

    @property
    def max_edge_length(self) -> Real:
        """Maximum edge length.

        The maximum facet edge length.
        """
        return self._max_edge_length

    @property
    def watertight(self) -> bool:
        """Watertight.

        Whether triangles on opposite sides of an edge should match.
        """
        return self._watertight



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing designs."""

from enum import Enum, unique
from pathlib import Path
from typing import Union

from beartype import beartype as check_input_types
from google.protobuf.empty_pb2 import Empty
import numpy as np
from pint import Quantity, UndefinedUnitError

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.commands_pb2 import (
    AssignMidSurfaceOffsetTypeRequest,
    AssignMidSurfaceThicknessRequest,
    CreateBeamCircularProfileRequest,
)
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from ansys.geometry.core.connection.backend import BackendType
from ansys.geometry.core.connection.conversions import (
    grpc_curve_to_curve,
    grpc_frame_to_frame,
    grpc_material_to_material,
    grpc_matrix_to_matrix,
    grpc_point_to_point3d,
    plane_to_grpc_plane,
    point3d_to_grpc_point,
)
from ansys.geometry.core.designer.beam import (
    Beam,
    BeamCircularProfile,
    BeamCrossSectionInfo,
    BeamProfile,
    BeamProperties,
    SectionAnchorType,
)
from ansys.geometry.core.designer.body import Body, MasterBody, MidSurfaceOffsetType
from ansys.geometry.core.designer.component import Component, SharedTopologyType
from ansys.geometry.core.designer.coordinate_system import CoordinateSystem
from ansys.geometry.core.designer.designpoint import DesignPoint
from ansys.geometry.core.designer.edge import Edge
from ansys.geometry.core.designer.face import Face
from ansys.geometry.core.designer.part import MasterComponent, Part
from ansys.geometry.core.designer.selection import NamedSelection
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.materials.material import Material
from ansys.geometry.core.materials.property import MaterialProperty, MaterialPropertyType
from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Y, ZERO_POINT3D
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.checks import ensure_design_is_active, min_backend_version
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Distance
from ansys.geometry.core.misc.options import ImportOptions
from ansys.geometry.core.modeler import Modeler
from ansys.geometry.core.parameters.parameter import Parameter, ParameterUpdateStatus
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.shapes.parameterization import Interval, ParamUV
from ansys.geometry.core.typing import RealSequence


@unique
class DesignFileFormat(Enum):
    """Provides supported file formats that can be downloaded for designs."""

    SCDOCX = "SCDOCX"
    PARASOLID_TEXT = "PARASOLID_TEXT"
    PARASOLID_BIN = "PARASOLID_BIN"
    FMD = "FMD"
    STEP = "STEP"
    IGES = "IGES"
    PMDB = "PMDB"
    STRIDE = "STRIDE"
    DISCO = "DISCO"
    INVALID = "INVALID"

    def __str__(self):
        """Represent object in string format."""
        return self.value


class Design(Component):
    """Provides for organizing geometry assemblies.

    This class synchronizes to a supporting Geometry service instance.

    Parameters
    ----------
    name : str
        User-defined label for the design.
    grpc_client : GrpcClient
        Active supporting Geometry service instance for design modeling.
    read_existing_design : bool, default: False
        Whether an existing design on the service should be read. This parameter is
        only valid when connecting to an existing service session. Otherwise, avoid
        using this optional parameter.
    """

    # Types of the class instance private attributes
    _materials: list[Material]
    _named_selections: dict[str, NamedSelection]
    _beam_profiles: dict[str, BeamProfile]

    @protect_grpc
    @check_input_types
    def __init__(self, name: str, modeler: Modeler, read_existing_design: bool = False):
        """Initialize the ``Design`` class."""
        super().__init__(name, None, modeler.client)

        # Initialize the stubs needed
        self._commands_stub = CommandsStub(self._grpc_client.channel)

        # Initialize needed instance variables
        self._materials = []
        self._named_selections = {}
        self._beam_profiles = {}
        self._design_id = ""
        self._is_active = False
        self._modeler = modeler

        # Check whether we want to process an existing design or create a new one.
        if read_existing_design:
            self._grpc_client.log.debug("Reading Design object from service.")
            self.__read_existing_design()
        else:
            response = self._grpc_client.services.designs.new(name=name)
            self._design_id = response.get("design_id")
            self._id = response.get("main_part_id")
            self._activate(called_after_design_creation=True)
            self._grpc_client.log.debug("Design object instantiated successfully.")

    @property
    def design_id(self) -> str:
        """The design's object unique id."""
        return self._design_id

    @property
    def materials(self) -> list[Material]:
        """List of materials available for the design."""
        return self._materials

    @property
    def named_selections(self) -> list[NamedSelection]:
        """List of named selections available for the design."""
        return list(self._named_selections.values())

    @property
    def beam_profiles(self) -> list[BeamProfile]:
        """List of beam profile available for the design."""
        return list(self._beam_profiles.values())

    @property
    def parameters(self) -> list[Parameter]:
        """List of parameters available for the design."""
        return self.get_all_parameters()

    @property
    def is_active(self) -> bool:
        """Whether the design is currently active."""
        return self._is_active

    @property
    def is_closed(self) -> bool:
        """Whether the design is closed (i.e. not active)."""
        return not self._is_active

    def close(self) -> None:
        """Close the design."""
        # Check if the design is already closed
        if self.is_closed:
            self._grpc_client.log.warning(f"Design {self.name} is already closed.")
            return

        # Attempt to close the design
        try:
            self._grpc_client.services.designs.close(design_id=self._design_id)
        except Exception as err:
            self._grpc_client.log.warning(f"Design {self.name} could not be closed. Error: {err}.")
            self._grpc_client.log.warning("Ignoring response and assuming the design is closed.")

        # Consider the design closed (even if the close request failed)
        self._is_active = False

    def _activate(self, called_after_design_creation: bool = False) -> None:
        """Activate the design."""
        # Activate the current design
        if not called_after_design_creation:
            self._grpc_client.services.designs.put_active(design_id=self._design_id)
        self._is_active = True
        self._grpc_client.log.debug(f"Design {self.name} is activated.")

    # TODO: allow for list of materials
    # https://github.com/ansys/pyansys-geometry/issues/1319
    @check_input_types
    @ensure_design_is_active
    def add_material(self, material: Material) -> None:
        """Add a material to the design.

        Parameters
        ----------
        material : Material
            Material to add.
        """
        self._grpc_client.services.materials.add_material(material=material)
        self._materials.append(material)
        self._grpc_client.log.debug(f"Material {material.name} is successfully added to design.")

    @check_input_types
    @ensure_design_is_active
    def save(self, file_location: Path | str) -> None:
        """Save a design to disk on the active Geometry server instance.

        Parameters
        ----------
        file_location : ~pathlib.Path | str
            Location on disk to save the file to.
        """
        # Sanity checks on inputs
        if isinstance(file_location, Path):
            file_location = str(file_location)

        self._grpc_client.services.designs.save_as(filepath=file_location)
        self._grpc_client.log.debug(f"Design successfully saved at location {file_location}.")

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def download(
        self,
        file_location: Path | str,
        format: DesignFileFormat = DesignFileFormat.SCDOCX,
    ) -> None:
        """Export and download the design from the server.

        Parameters
        ----------
        file_location : ~pathlib.Path | str
            Location on disk to save the file to.
        format : DesignFileFormat, default: DesignFileFormat.SCDOCX
            Format for the file to save to.
        """
        # Sanity checks on inputs
        if isinstance(file_location, str):
            file_location = Path(file_location)

        # Check if the folder for the file location exists
        if not file_location.parent.exists():
            # Create the parent directory
            file_location.parent.mkdir(parents=True, exist_ok=True)

        # Process response
        self._grpc_client.log.debug(f"Requesting design download in {format} format.")
        if self._modeler.client.backend_version < (25, 2, 0):
            received_bytes = self.__export_and_download_legacy(format=format)
        else:
            received_bytes = self.__export_and_download(format=format)

        # Write to file
        file_location.write_bytes(received_bytes)
        self._grpc_client.log.debug(f"Design downloaded at location {file_location}.")

    def __export_and_download_legacy(self, format: DesignFileFormat) -> bytes:
        """Export and download the design from the server.

        Parameters
        ----------
        format : DesignFileFormat
            Format for the file to save to.

        Returns
        -------
        bytes
            The raw data from the exported and downloaded file.

        Notes
        -----
        This is a legacy method, which is used in versions
        up to Ansys 25.1.1 products.
        """
        # Process response
        self._grpc_client.log.debug(f"Requesting design download in {format} format.")
        if format is DesignFileFormat.SCDOCX:
            response = self._commands_stub.DownloadFile(Empty())
            received_bytes = bytes()
            received_bytes += response.data
        elif format in [
            DesignFileFormat.PARASOLID_TEXT,
            DesignFileFormat.PARASOLID_BIN,
            DesignFileFormat.FMD,
            DesignFileFormat.STEP,
            DesignFileFormat.IGES,
            DesignFileFormat.PMDB,
        ]:
            response = self._grpc_client.services.parts.export(format=format)
            received_bytes = response.get("data")
        else:
            self._grpc_client.log.warning(
                f"{format} format requested is not supported. Ignoring download request."
            )
            return

        return received_bytes

    def __export_and_download(self, format: DesignFileFormat) -> bytes:
        """Export and download the design from the server.

        Parameters
        ----------
        format : DesignFileFormat
            Format for the file to save to.

        Returns
        -------
        bytes
            The raw data from the exported and downloaded file.
        """
        # Process response
        self._grpc_client.log.debug(f"Requesting design download in {format} format.")

        if format in [
            DesignFileFormat.PARASOLID_TEXT,
            DesignFileFormat.PARASOLID_BIN,
            DesignFileFormat.FMD,
            DesignFileFormat.STEP,
            DesignFileFormat.IGES,
            DesignFileFormat.PMDB,
            DesignFileFormat.DISCO,
            DesignFileFormat.SCDOCX,
            DesignFileFormat.STRIDE,
        ]:
            try:
                response = self._grpc_client.services.designs.download_export(format=format)
            except Exception:
                self._grpc_client.log.warning(
                    f"Failed to download the file in {format} format."
                    " Attempting to stream download."
                )
                # Attempt to download the file via streaming
                response = self._grpc_client.services.designs.stream_download_export(format=format)
        else:
            self._grpc_client.log.warning(
                f"{format} format requested is not supported. Ignoring download request."
            )
            return None

        return response.get("data")

    def __build_export_file_location(self, location: Path | str | None, ext: str) -> Path:
        """Build the file location for export functions.

        Parameters
        ----------
        location : ~pathlib.Path | str
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.
        ext : str
            Extension to use for the file.

        Returns
        -------
        ~pathlib.Path
            The file location for the export function.
        """
        return (Path(location) if location else Path.cwd()) / f"{self.name}.{ext}"

    def export_to_scdocx(self, location: Path | str | None = None) -> Path:
        """Export the design to an scdocx file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "scdocx")

        # Export the design to an scdocx file
        self.download(file_location, DesignFileFormat.SCDOCX)

        # Return the file location
        return file_location

    def export_to_disco(self, location: Path | str | None = None) -> Path:
        """Export the design to an dsco file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "dsco")

        # Export the design to an dsco file
        self.download(file_location, DesignFileFormat.DISCO)

        # Return the file location
        return file_location

    def export_to_stride(self, location: Path | str | None = None) -> Path:
        """Export the design to an stride file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "stride")

        # Export the design to an stride file
        self.download(file_location, DesignFileFormat.STRIDE)

        # Return the file location
        return file_location

    def export_to_parasolid_text(self, location: Path | str | None = None) -> Path:
        """Export the design to a Parasolid text file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Determine the extension based on the backend type
        ext = "xmt_txt" if BackendType.is_linux_service(self._grpc_client.backend_type) else "x_t"

        # Define the file location
        file_location = self.__build_export_file_location(location, ext)

        # Export the design to a Parasolid text file
        self.download(file_location, DesignFileFormat.PARASOLID_TEXT)

        # Return the file location
        return file_location

    def export_to_parasolid_bin(self, location: Path | str | None = None) -> Path:
        """Export the design to a Parasolid binary file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Determine the extension based on the backend type
        ext = "xmt_bin" if BackendType.is_linux_service(self._grpc_client.backend_type) else "x_b"

        # Define the file location
        file_location = self.__build_export_file_location(location, ext)

        # Export the design to a Parasolid binary file
        self.download(file_location, DesignFileFormat.PARASOLID_BIN)

        # Return the file location
        return file_location

    def export_to_fmd(self, location: Path | str | None = None) -> Path:
        """Export the design to an FMD file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "fmd")

        # Export the design to an FMD file
        self.download(file_location, DesignFileFormat.FMD)

        # Return the file location
        return file_location

    def export_to_step(self, location: Path | str | None = None) -> Path:
        """Export the design to a STEP file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "stp")

        # Export the design to a STEP file
        self.download(file_location, DesignFileFormat.STEP)

        # Return the file location
        return file_location

    def export_to_iges(self, location: Path | str = None) -> Path:
        """Export the design to an IGES file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "igs")

        # Export the design to an IGES file
        self.download(file_location, DesignFileFormat.IGES)

        # Return the file location
        return file_location

    def export_to_pmdb(self, location: Path | str | None = None) -> Path:
        """Export the design to a PMDB file.

        Parameters
        ----------
        location : ~pathlib.Path | str, optional
            Location on disk to save the file to. If None, the file will be saved
            in the current working directory.

        Returns
        -------
        ~pathlib.Path
            The path to the saved file.
        """
        # Define the file location
        file_location = self.__build_export_file_location(location, "pmdb")

        # Export the design to a PMDB file
        self.download(file_location, DesignFileFormat.PMDB)

        # Return the file location
        return file_location

    @check_input_types
    @ensure_design_is_active
    def create_named_selection(
        self,
        name: str,
        bodies: list[Body] | None = None,
        faces: list[Face] | None = None,
        edges: list[Edge] | None = None,
        beams: list[Beam] | None = None,
        design_points: list[DesignPoint] | None = None,
    ) -> NamedSelection:
        """Create a named selection on the active Geometry server instance.

        Parameters
        ----------
        name : str
            User-defined name for the named selection.
        bodies : list[Body], default: None
            All bodies to include in the named selection.
        faces : list[Face], default: None
            All faces to include in the named selection.
        edges : list[Edge], default: None
            All edges to include in the named selection.
        beams : list[Beam], default: None
            All beams to include in the named selection.
        design_points : list[DesignPoint], default: None
            All design points to include in the named selection.

        Returns
        -------
        NamedSelection
            Newly created named selection that maintains references to all target entities.
        """
        named_selection = NamedSelection(
            name,
            self,
            self._grpc_client,
            bodies=bodies,
            faces=faces,
            edges=edges,
            beams=beams,
            design_points=design_points,
        )

        self._named_selections[named_selection.name] = named_selection
        self._grpc_client.log.debug(
            f"Named selection {named_selection.name} is successfully created."
        )

        return self._named_selections[named_selection.name]

    @check_input_types
    @ensure_design_is_active
    def delete_named_selection(self, named_selection: NamedSelection | str) -> None:
        """Delete a named selection on the active Geometry server instance.

        Parameters
        ----------
        named_selection : NamedSelection | str
            Name of the named selection or instance.
        """
        if isinstance(named_selection, str):
            removal_name = named_selection
            removal = self._named_selections.get(named_selection, None)
            removal_id = removal.id if removal else None
        else:
            removal_name = named_selection.name
            removal_id = named_selection.id

        self._grpc_client.log.debug(f"Named selection {removal_name} deletion request received.")
        self._grpc_client.services.named_selection.delete_named_selection(id=removal_id)

        try:
            self._named_selections.pop(removal_name)
            self._grpc_client.log.debug(f"Named selection {removal_name} is successfully deleted.")
        except KeyError:
            self._grpc_client.log.warning(
                f"Attempted named selection deletion failed, with name {removal_name}."
                + " Ignoring this request."
            )
            pass

    @check_input_types
    @ensure_design_is_active
    def delete_component(self, component: Union["Component", str]) -> None:
        """Delete a component (itself or its children).

        Parameters
        ----------
        id : Union[Component, str]
            Name of the component or instance to delete.

        Raises
        ------
        ValueError
            The design itself cannot be deleted.

        Notes
        -----
        If the component is not this component (or its children), it
        is not deleted.
        """
        id = component if isinstance(component, str) else component.id
        if id == self.id:
            raise ValueError("The design itself cannot be deleted.")
        else:
            return super().delete_component(component)

    def set_shared_topology(self, share_type: SharedTopologyType) -> None:
        """Set the shared topology to apply to the component.

        Parameters
        ----------
        share_type : SharedTopologyType
            Shared topology type to assign.

        Raises
        ------
        ValueError
            Shared topology does not apply to a design.
        """
        raise ValueError("The design itself cannot have a shared topology.")

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def add_beam_circular_profile(
        self,
        name: str,
        radius: Quantity | Distance,
        center: np.ndarray | RealSequence | Point3D = ZERO_POINT3D,
        direction_x: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_X,
        direction_y: np.ndarray | RealSequence | UnitVector3D | Vector3D = UNITVECTOR3D_Y,
    ) -> BeamCircularProfile:
        """Add a new beam circular profile under the design for creating beams.

        Parameters
        ----------
        name : str
            User-defined label for the new beam circular profile.
        radius : ~pint.Quantity | Distance
            Radius of the beam circular profile.
        center : ~numpy.ndarray | RealSequence | Point3D
            Center of the beam circular profile.
        direction_x : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
            X-plane direction.
        direction_y : ~numpy.ndarray | RealSequence | UnitVector3D | Vector3D
            Y-plane direction.
        """
        dir_x = direction_x if isinstance(direction_x, UnitVector3D) else UnitVector3D(direction_x)
        dir_y = direction_y if isinstance(direction_y, UnitVector3D) else UnitVector3D(direction_y)
        radius = radius if isinstance(radius, Distance) else Distance(radius)

        if radius.value <= 0:
            raise ValueError("Radius must be a real positive value.")

        if not dir_x.is_perpendicular_to(dir_y):
            raise ValueError("Direction X and direction Y must be perpendicular.")

        request = CreateBeamCircularProfileRequest(
            origin=point3d_to_grpc_point(center),
            radius=radius.value.m_as(DEFAULT_UNITS.SERVER_LENGTH),
            plane=plane_to_grpc_plane(Plane(center, dir_x, dir_y)),
            name=name,
        )

        self._grpc_client.log.debug(f"Creating a beam circular profile on {self.id}...")

        response = self._commands_stub.CreateBeamCircularProfile(request)
        profile = BeamCircularProfile(response.id, name, radius, center, dir_x, dir_y)
        self._beam_profiles[profile.name] = profile

        self._grpc_client.log.debug(
            f"Beam circular profile {profile.name} is successfully created."
        )

        return self._beam_profiles[profile.name]

    @min_backend_version(25, 1, 0)
    def get_all_parameters(self) -> list[Parameter]:
        """Get parameters for the design.

        Returns
        -------
        list[Parameter]
            List of parameters for the design.
        """
        response = self._grpc_client._services.driving_dimensions.get_all_parameters()
        return response.get("parameters")

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_parameter(self, dimension: Parameter) -> ParameterUpdateStatus:
        """Set or update a parameter of the design.

        Parameters
        ----------
        dimension : Parameter
            Parameter to set.

        Returns
        -------
        ParameterUpdateStatus
            Status of the update operation.
        """
        response = self._grpc_client._services.driving_dimensions.set_parameter(
            driving_dimension=dimension
        )

        # Update the design in place. This method is computationally expensive,
        # consider finding a more efficient approach.
        self._update_design_inplace()

        return response.get("status")

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def add_midsurface_thickness(self, thickness: Quantity, bodies: list[Body]) -> None:
        """Add a mid-surface thickness to a list of bodies.

        Parameters
        ----------
        thickness : ~pint.Quantity
            Thickness to be assigned.
        bodies : list[Body]
            All bodies to include in the mid-surface thickness assignment.

        Notes
        -----
        Only surface bodies will be eligible for mid-surface thickness assignment.
        """
        # Store only assignable ids
        ids: list[str] = []
        ids_bodies: list[Body] = []
        for body in bodies:
            if body.is_surface:
                ids.append(body.id)
                ids_bodies.append(body)
            else:
                self._grpc_client.log.warning(
                    f"Body {body.name} cannot be assigned a mid-surface thickness since it is not a surface. Ignoring request."  # noqa : E501
                )

        # Assign mid-surface thickness
        self._commands_stub.AssignMidSurfaceThickness(
            AssignMidSurfaceThicknessRequest(
                bodies_or_faces=ids, thickness=thickness.m_as(DEFAULT_UNITS.SERVER_LENGTH)
            )
        )

        # Once the assignment has gone fine, store the values
        for body in ids_bodies:
            body._surface_thickness = thickness

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def add_midsurface_offset(self, offset_type: MidSurfaceOffsetType, bodies: list[Body]) -> None:
        """Add a mid-surface offset type to a list of bodies.

        Parameters
        ----------
        offset_type : MidSurfaceOffsetType
            Surface offset to be assigned.
        bodies : list[Body]
            All bodies to include in the mid-surface offset assignment.

        Notes
        -----
        Only surface bodies will be eligible for mid-surface offset assignment.
        """
        # Store only assignable ids
        ids: list[str] = []
        ids_bodies: list[Body] = []
        for body in bodies:
            if body.is_surface:
                ids.append(body.id)
                ids_bodies.append(body)
            else:
                self._grpc_client.log.warning(
                    f"Body {body.name} cannot be assigned a mid-surface offset since it is not a surface. Ignoring request."  # noqa : E501
                )

        # Assign mid-surface offset type
        self._commands_stub.AssignMidSurfaceOffsetType(
            AssignMidSurfaceOffsetTypeRequest(bodies_or_faces=ids, offset_type=offset_type.value)
        )

        # Once the assignment has gone fine, store the values
        for body in ids_bodies:
            body._surface_offset = offset_type

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    def delete_beam_profile(self, beam_profile: BeamProfile | str) -> None:
        """Remove a beam profile on the active geometry server instance.

        Parameters
        ----------
        beam_profile : BeamProfile | str
            A beam profile name or instance that should be deleted.
        """
        removal_name = beam_profile if isinstance(beam_profile, str) else beam_profile.name
        self._grpc_client.log.debug(f"Beam profile {removal_name} deletion request received.")
        removal_obj = self._beam_profiles.get(removal_name, None)

        if removal_obj:
            self._commands_stub.DeleteBeamProfile(EntityIdentifier(id=removal_obj.id))
            self._beam_profiles.pop(removal_name)
            self._grpc_client.log.debug(f"Beam profile {removal_name} successfully deleted.")
        else:
            self._grpc_client.log.warning(
                f"Attempted beam profile deletion failed, with name {removal_name}."
                + " Ignoring request."
            )

    @protect_grpc
    @check_input_types
    @ensure_design_is_active
    @min_backend_version(24, 2, 0)
    def insert_file(
        self, file_location: Path | str, import_options: ImportOptions = ImportOptions()
    ) -> Component:
        """Insert a file into the design.

        Parameters
        ----------
        file_location : ~pathlib.Path | str
            Location on disk where the file is located.
        import_options : ImportOptions
            The options to pass into upload file

        Returns
        -------
        Component
            The newly inserted component.
        """
        # Upload the file to the server
        filepath_server = self._modeler._upload_file(file_location, import_options=import_options)

        # Insert the file into the design
        self._grpc_client.services.designs.insert(filepath=filepath_server)
        self._grpc_client.log.debug(f"File {file_location} successfully inserted into design.")

        self._update_design_inplace()

        self._grpc_client.log.debug(f"Design {self.name} is successfully updated.")

        # Return the newly inserted component
        return self._components[-1]

    def __repr__(self) -> str:
        """Represent the ``Design`` as a string."""
        alive_bodies = [1 if body.is_alive else 0 for body in self.bodies]
        alive_comps = [1 if comp.is_alive else 0 for comp in self.components]
        lines = [f"ansys.geometry.core.designer.Design {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Is active?           : {self._is_active}")
        lines.append(f"  N Bodies             : {sum(alive_bodies)}")
        lines.append(f"  N Components         : {sum(alive_comps)}")
        lines.append(f"  N Coordinate Systems : {len(self.coordinate_systems)}")
        lines.append(f"  N Named Selections   : {len(self.named_selections)}")
        lines.append(f"  N Materials          : {len(self.materials)}")
        lines.append(f"  N Beam Profiles      : {len(self.beam_profiles)}")
        lines.append(f"  N Design Points      : {len(self.design_points)}")
        return "\n".join(lines)

    def __read_existing_design(self) -> None:
        """Read an existing ``Design`` located on the server."""
        #
        # This might go out of sync with the _update_design_inplace method.
        # Ensure that the two methods are in sync. Especially regarding cleanup.
        #
        # TODO: Not all features implemented yet. Status is as follows
        #
        # Windows:
        #
        # - [X] Components
        # - [X] Bodies
        # - [X] Materials
        # - [X] NamedSelections
        # - [ ] BeamProfiles
        # - [ ] Beams
        # - [X] CoordinateSystems
        # - [X] SharedTopology
        #
        # Linux:
        #
        # - [X] Components
        # - [X] Bodies
        # - [X] Materials
        # - [X] NamedSelections
        # - [ ] BeamProfiles
        # - [ ] Beams
        # - [X] CoordinateSystems
        # - [ ] SharedTopology
        #
        # https://github.com/ansys/pyansys-geometry/issues/1319
        import time

        start = time.time()
        # Grab active design
        design_response = self._grpc_client.services.designs.get_active()
        if not design_response:
            raise RuntimeError("No existing design available at service level.")
        else:
            self._design_id = design_response.get("design_id")
            self._id = design_response.get("main_part_id")
            self._name = design_response.get("name")
            self._activate(called_after_design_creation=True)

        response = self._commands_stub.GetAssembly(EntityIdentifier(id=self._design_id))

        # Store created objects
        created_parts = {p.id: Part(p.id, p.name, [], []) for p in response.parts}
        created_tps = {}
        created_components = {design_response.get("main_part_id"): self}
        created_bodies = {}

        # Make dummy master for design since server doesn't have one
        self._master_component = MasterComponent(
            "1", "master_design", created_parts[design_response.get("main_part_id")]
        )

        # Create MasterComponents
        for master in response.transformed_parts:
            part = created_parts.get(master.part_master.id)
            new_master = MasterComponent(
                master.id, master.name, part, grpc_matrix_to_matrix(master.placement)
            )
            created_tps[master.id] = new_master

        # Create Components
        for comp in response.components:
            parent = created_components.get(comp.parent_id)
            master = created_tps.get(comp.master_id)
            c = Component(
                comp.name,
                parent,
                self._grpc_client,
                preexisting_id=comp.id,
                master_component=master,
                read_existing_comp=True,
            )
            created_components[comp.id] = c
            parent.components.append(c)

        # Create Bodies
        for body in response.bodies:
            part = created_parts.get(body.parent_id)
            tb = MasterBody(body.id, body.name, self._grpc_client, is_surface=body.is_surface)
            part.bodies.append(tb)
            created_bodies[body.id] = tb

        # Create Materials
        for material in response.materials:
            properties = []
            density = Quantity(0)
            for property in material.material_properties:
                # TODO: Add support for more material properties...
                #      - Need to add support for more MaterialPropertyTypes
                #      - Need to add support for more Quantity units
                # https://github.com/ansys/pyansys-geometry/issues/1319
                try:
                    mp_type = MaterialPropertyType.from_id(property.id)
                except ValueError as err:
                    # TODO: Errors coming from MaterialPropertyType.from_id
                    # because of unsupported MaterialPropertyType entries...
                    # https://github.com/ansys/pyansys-geometry/issues/1319
                    self._grpc_client.log.warning(
                        f"Material property {property.display_name} of type {property.id} is not supported."  # noqa : E501
                        " Storing as string."
                    )
                    self._grpc_client.log.warning(f"Root cause: {err}")
                    mp_type = property.id

                try:
                    mp_quantity = Quantity(property.value, property.units)
                except (
                    UndefinedUnitError,
                    TypeError,
                ) as err:  # TODO: Errors coming from Quantity ctor because of unsupported units...
                    # https://github.com/ansys/pyansys-geometry/issues/1319
                    self._grpc_client.log.warning(
                        f"Material property {property.display_name} with units {property.units} is not fully supported."  # noqa : E501
                        " Storing value only as float."
                    )
                    self._grpc_client.log.warning(f"Root cause: {err}")
                    mp_quantity = property.value

                mp = MaterialProperty(mp_type, property.display_name, mp_quantity)
                properties.append(mp)
                if mp.type == MaterialPropertyType.DENSITY:
                    density = mp.quantity

            m = Material(material.name, density, properties)
            self.materials.append(m)

        # Create Beams
        for beam in response.beams:
            cross_section = BeamCrossSectionInfo(
                section_anchor=SectionAnchorType(beam.cross_section.section_anchor),
                section_angle=beam.cross_section.section_angle,
                section_frame=grpc_frame_to_frame(beam.cross_section.section_frame),
                section_profile=[
                    [
                        TrimmedCurve(
                            geometry=grpc_curve_to_curve(curve.curve),
                            start=grpc_point_to_point3d(curve.start),
                            end=grpc_point_to_point3d(curve.end),
                            interval=Interval(curve.interval_start, curve.interval_end),
                            length=curve.length,
                        )
                        for curve in curve_list.curves
                    ]
                    for curve_list in beam.cross_section.section_profile
                ],
            )
            properties = BeamProperties(
                area=beam.properties.area,
                centroid=ParamUV(beam.properties.centroid_x, beam.properties.centroid_y),
                warping_constant=beam.properties.warping_constant,
                ixx=beam.properties.ixx,
                ixy=beam.properties.ixy,
                iyy=beam.properties.iyy,
                shear_center=ParamUV(
                    beam.properties.shear_center_x, beam.properties.shear_center_y
                ),
                torsion_constant=beam.properties.torsional_constant,
            )

            new_beam = Beam(
                id=beam.id.id,
                start=grpc_point_to_point3d(beam.shape.start),
                end=grpc_point_to_point3d(beam.shape.end),
                profile=None,
                # TODO: Beams need BeamProfiles imported from existing design
                # https://github.com/ansys/pyansys-geometry/issues/1825
                parent_component=self,
                name=beam.name,
                is_deleted=beam.is_deleted,
                is_reversed=beam.is_reversed,
                is_rigid=beam.is_rigid,
                material=grpc_material_to_material(beam.material),
                cross_section=cross_section,
                properties=properties,
                shape=beam.shape,
                beam_type=beam.type,
            )

            # Find the component to which the beam belongs
            parent = created_components.get(beam.parent.id, self)
            parent._beams.append(new_beam)

        # Create NamedSelections
        for ns in response.named_selections:
            new_ns = NamedSelection(
                ns.name,
                self,
                self._grpc_client,
                preexisting_id=ns.id,
            )
            self._named_selections[new_ns.name] = new_ns

        # Create CoordinateSystems
        num_created_coord_systems = 0
        for component_id, coordinate_systems in response.component_coord_systems.items():
            component = created_components.get(component_id)
            for cs in coordinate_systems.coordinate_systems:
                frame = grpc_frame_to_frame(cs.frame)
                new_cs = CoordinateSystem(cs.name, frame, component, self._grpc_client, cs.id)
                component.coordinate_systems.append(new_cs)
                num_created_coord_systems += 1

        end = time.time()

        # Set SharedTopology
        # TODO: Maybe just add it to Component or Part message
        # we're starting to iterate through all the Components too much.
        # Make sure design doesn't need edge case attention
        # https://github.com/ansys/pyansys-geometry/issues/1319
        num_created_shared_topologies = 0
        for component_id, shared_topology_type in response.component_shared_topologies.items():
            component = created_components.get(component_id)
            component._shared_topology = SharedTopologyType(shared_topology_type)
            num_created_shared_topologies += 1

        self._grpc_client.log.debug(f"Parts created: {len(created_parts)}")
        self._grpc_client.log.debug(f"MasterComponents created: {len(created_tps) + 1}")
        self._grpc_client.log.debug(f"Components created: {len(created_components)}")
        self._grpc_client.log.debug(f"Bodies created: {len(created_bodies)}")
        self._grpc_client.log.debug(f"Materials created: {len(self.materials)}")
        self._grpc_client.log.debug(f"NamedSelections created: {len(self.named_selections)}")
        self._grpc_client.log.debug(f"CoordinateSystems created: {num_created_coord_systems}")
        self._grpc_client.log.debug(f"SharedTopologyTypes set: {num_created_shared_topologies}")

        self._grpc_client.log.debug(f"\nSuccessfully read design in: {end - start} s")

    def _update_design_inplace(self) -> None:
        """Update the design to align with the server side.

        Notes
        -----
        This method is used to update the design inside repair tools.
        Its usage is not recommended for other purposes.
        """
        # Clear all the existing information
        #
        # TODO: This might go out of sync with the __read_existing_design method
        # if the latter is updated and this method is not. Ensure that
        # the two methods are in sync.
        # https://github.com/ansys/pyansys-geometry/issues/1319
        #
        self._components = []
        self._clear_cached_bodies()
        self._materials = []
        self._named_selections = {}
        self._coordinate_systems = {}

        # Read the existing design
        self.__read_existing_design()



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Unsupported functions for the PyAnsys Geometry library."""

from enum import Enum, unique
from typing import TYPE_CHECKING

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.unsupported_pb2 import ExportIdRequest, ImportIdRequest
from ansys.api.geometry.v0.unsupported_pb2_grpc import UnsupportedStub
from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.misc.auxiliary import get_all_bodies_from_design
from ansys.geometry.core.misc.checks import (
    min_backend_version,
)

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face
    from ansys.geometry.core.modeler import Modeler


@unique
class PersistentIdType(Enum):
    """Type of persistent id."""

    PNAME = 1
    PRIME_ID = 700


class UnsupportedCommands:
    """Provides unsupported commands for PyAnsys Geometry.

    Parameters
    ----------
    grpc_client : GrpcClient
        gRPC client to use for the geometry commands.
    modeler : Modeler
        Modeler instance to use for the geometry commands.
    """

    def __init__(self, grpc_client: GrpcClient, modeler: "Modeler"):
        """Initialize an instance of the ``UnsupportedCommands`` class."""
        self._grpc_client = grpc_client
        self._unsupported_stub = UnsupportedStub(self._grpc_client.channel)
        self.__id_map = {}
        self.__modeler = modeler
        self.__current_design = modeler.get_active_design()

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def __fill_imported_id_map(self, id_type: PersistentIdType) -> None:
        """Populate the persistent id map for caching.

        Parameters
        ----------
        id_type : PersistentIdType
            Type of id.

        Notes
        -----
        This cache should be cleared on design change.
        """
        request = ImportIdRequest(type=id_type.value)
        self.__id_map[id_type] = self._unsupported_stub.GetImportIdMap(request).id_map

    def __clear_cache(self) -> None:
        """Clear the cache of persistent id's.

        Notes
        -----
        This should be called on design change.
        """
        self.__id_map = {}

    def __is_occurrence(self, master: EntityIdentifier, occ: str) -> bool:
        """Determine if the master is the master of the occurrence.

        Parameters
        ----------
        master : EntityIdentifier
            Master moniker.
        occ : str
            Occurrence moniker.

        Returns
        -------
        bool
            ``True`` if the master is the master of the occurrence.

        """
        master_id = occ.split("/")[-1]
        return master.id == master_id

    def __get_moniker_from_import_id(
        self, id_type: PersistentIdType, import_id: str
    ) -> EntityIdentifier | None:
        """Look up the moniker from the id map.

        Parameters
        ----------
        id_type : PersistentIdType
            Type of id.
        import_id : str
            Persistent id.

        Returns
        -------
        EntityIdentifier
            Moniker associated with the id or None

        Notes
        -----
        This checks if the design has changed and clears the cache if it has.
        """
        if (
            self.__current_design != self.__modeler.get_active_design()
            or id_type not in self.__id_map
        ):
            self.__fill_imported_id_map(id_type)
        moniker = self.__id_map[id_type].get(import_id, None)
        return moniker

    @protect_grpc
    @min_backend_version(25, 2, 0)
    def set_export_id(self, moniker: str, id_type: PersistentIdType, value: str) -> None:
        """Set the persistent id for the moniker.

        Parameters
        ----------
        moniker : str
            Moniker to set the id for.
        id_type : PersistentIdType
            Type of id.
        value : str
            Id to set.
        """
        request = ExportIdRequest(
            moniker=EntityIdentifier(id=moniker), id=value, type=id_type.value
        )
        self._unsupported_stub.SetExportId(request)
        self.__id_map = {}

    def get_body_occurrences_from_import_id(
        self, import_id: str, id_type: PersistentIdType
    ) -> list["Body"]:
        """Get all body occurrences whose master has the given import id.

        Parameters
        ----------
        import_id : str
            Persistent id
        id_type : PersistentIdType
            Type of id

        Returns
        -------
        list[Body]
            List of body occurrences.
        """
        moniker = self.__get_moniker_from_import_id(id_type, import_id)

        if moniker is None:
            return []

        design = self.__modeler.get_active_design()
        return [
            body
            for body in get_all_bodies_from_design(design)
            if self.__is_occurrence(moniker, body.id)
        ]

    def get_face_occurrences_from_import_id(
        self, import_id: str, id_type: PersistentIdType
    ) -> list["Face"]:
        """Get all face occurrences whose master has the given import id.

        Parameters
        ----------
        import_id : str
            Persistent id.
        id_type : PersistentIdType
            Type of id.

        Returns
        -------
        list[Face]
            List of face occurrences.
        """
        moniker = self.__get_moniker_from_import_id(id_type, import_id)

        if moniker is None:
            return []

        design = self.__modeler.get_active_design()
        return [
            face
            for body in get_all_bodies_from_design(design)
            for face in body.faces
            if self.__is_occurrence(moniker, face.id)
        ]

    def get_edge_occurrences_from_import_id(
        self, import_id: str, id_type: PersistentIdType
    ) -> list["Edge"]:
        """Get all edge occurrences whose master has the given import id.

        Parameters
        ----------
        import_id : str
           Persistent id.
        id_type : PersistentIdType
           Type of id.

        Returns
        -------
        list[Edge]
           List of edge occurrences.
        """
        moniker = self.__get_moniker_from_import_id(id_type, import_id)

        if moniker is None:
            return []

        design = self.__modeler.get_active_design()
        return [
            edge
            for body in get_all_bodies_from_design(design)
            for edge in body.edges
            if self.__is_occurrence(moniker, edge.id)
        ]



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides get and set methods for parameters."""

from enum import Enum, unique

from ansys.geometry.core.typing import Real


@unique
class ParameterType(Enum):
    """Provides values for the parameter types supported."""

    DIMENSIONTYPE_UNKNOWN = 0
    DIMENSIONTYPE_LINEAR = 1
    DIMENSIONTYPE_DIAMETRIC = 2
    DIMENSIONTYPE_RADIAL = 3
    DIMENSIONTYPE_ARC = 4
    DIMENSIONTYPE_AREA = 5
    DIMENSIONTYPE_VOLUME = 6
    DIMENSIONTYPE_MASS = 7
    DIMENSIONTYPE_ANGULAR = 8
    DIMENSIONTYPE_COUNT = 9
    DIMENSIONTYPE_UNITLESS = 10


@unique
class ParameterUpdateStatus(Enum):
    """Provides values for the status messages associated with parameter updates."""

    SUCCESS = 0
    FAILURE = 1
    CONSTRAINED_PARAMETERS = 2
    UNKNOWN = 3


class Parameter:
    """Represents a parameter.

    Parameters
    ----------
    id : int
        Unique ID for the parameter.
    name : str
        Name of the parameter.
    dimension_type : ParameterType
        Type of the parameter.
    dimension_value : float
        Value of the parameter.
    """

    def __init__(self, id: int, name: str, dimension_type: ParameterType, dimension_value: Real):
        """Initialize an instance of the ``Parameter`` class."""
        self.id = id
        self._name = name
        self._dimension_type = dimension_type
        self._dimension_value = dimension_value

    @property
    def name(self) -> str:
        """Get the name of the parameter."""
        return self._name

    @name.setter
    def name(self, value: str):
        """Set the name of the parameter."""
        self._name = value

    @property
    def dimension_value(self) -> Real:
        """Get the value of the parameter."""
        return self._dimension_value

    @dimension_value.setter
    def dimension_value(self, value: Real):
        """Set the value of the parameter."""
        self._dimension_value = value

    @property
    def dimension_type(self) -> ParameterType:
        """Get the type of the parameter."""
        return self._dimension_type

    @dimension_type.setter
    def dimension_type(self, value: ParameterType):
        """Set the type of the parameter."""
        self._dimension_type = value



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a body."""

from abc import ABC, abstractmethod
from collections.abc import Iterable
from enum import Enum, unique
from functools import cached_property, wraps
from typing import TYPE_CHECKING, Union

from beartype import beartype as check_input_types
import matplotlib.colors as mcolors
from pint import Quantity

from ansys.api.dbu.v0.dbumodels_pb2 import EntityIdentifier
from ansys.api.geometry.v0.commands_pb2 import (
    AssignMidSurfaceOffsetTypeRequest,
    AssignMidSurfaceThicknessRequest,
    CombineIntersectBodiesRequest,
    CombineMergeBodiesRequest,
    ImprintCurvesRequest,
    ProjectCurvesRequest,
    RemoveFacesRequest,
    ShellRequest,
)
from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub
from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.connection.conversions import (
    plane_to_grpc_plane,
    sketch_shapes_to_grpc_geometries,
    trimmed_curve_to_grpc_trimmed_curve,
    unit_vector_to_grpc_direction,
)
from ansys.geometry.core.designer.edge import CurveType, Edge
from ansys.geometry.core.designer.face import Face, SurfaceType
from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.materials.material import Material
from ansys.geometry.core.math.bbox import BoundingBox
from ansys.geometry.core.math.constants import IDENTITY_MATRIX44
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.auxiliary import (
    DEFAULT_COLOR,
    convert_color_to_hex,
    convert_opacity_to_hex,
    get_design_from_body,
)
from ansys.geometry.core.misc.checks import (
    check_type,
    check_type_all_elements_in_iterable,
    ensure_design_is_active,
    graphics_required,
    min_backend_version,
)
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Angle, Distance
from ansys.geometry.core.misc.options import TessellationOptions
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.sketch.sketch import Sketch
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    from pyvista import MultiBlock, PolyData

    from ansys.geometry.core.designer.component import Component

# TODO: Temporary fix for boolean operations
# This is a temporary fix for the boolean operations issue. The issue is that the
# boolean operations are not working as expected with command-based operations. The
# goal is to fix this issue in the future.
# https://github.com/ansys/pyansys-geometry/issues/1733
__TEMPORARY_BOOL_OPS_FIX__ = (99, 0, 0)


@unique
class MidSurfaceOffsetType(Enum):
    """Provides values for mid-surface offsets supported."""

    MIDDLE = 0
    TOP = 1
    BOTTOM = 2
    VARIABLE = 3
    CUSTOM = 4


@unique
class CollisionType(Enum):
    """Provides values for collision types between bodies."""

    NONE = 0
    TOUCH = 1
    INTERSECT = 2
    CONTAINED = 3
    CONTAINEDTOUCH = 4


@unique
class FillStyle(Enum):
    """Provides values for fill styles supported."""

    DEFAULT = 0
    OPAQUE = 1
    TRANSPARENT = 2


class IBody(ABC):
    """Defines the common methods for a body, providing the abstract body interface.

    Both the ``MasterBody`` class and ``Body`` class both inherit from the ``IBody``
    class. All child classes must implement all abstract methods.
    """

    @abstractmethod
    def id(self) -> str:
        """Get the ID of the body as a string."""
        return

    @abstractmethod
    def _grpc_id(self) -> EntityIdentifier:
        """Entity identifier of this body on the server side."""
        return

    @abstractmethod
    def name(self) -> str:
        """Get the name of the body."""
        return

    @abstractmethod
    def set_name(self, str) -> None:
        """Set the name of the body."""
        return

    @abstractmethod
    def fill_style(self) -> FillStyle:
        """Get the fill style of the body."""
        return

    @abstractmethod
    def set_fill_style(self, fill_style: FillStyle) -> None:
        """Set the fill style of the body."""
        return

    @abstractmethod
    def is_suppressed(self) -> bool:
        """Get the body suppression state."""
        return

    @abstractmethod
    def set_suppressed(self, suppressed: bool) -> None:
        """Set the body suppression state."""
        return

    @abstractmethod
    def color(self) -> str:
        """Get the color of the body."""
        return

    @abstractmethod
    def opacity(self) -> float:
        """Get the float value of the opacity for the body."""
        return

    @abstractmethod
    def set_color(self, color: str | tuple[float, float, float]) -> None:
        """Set the color of the body.

        Parameters
        ----------
        color : str | tuple[float, float, float]
            Color to set the body to. This can be a string representing a color name
            or a tuple of RGB values in the range [0, 1] (RGBA) or [0, 255] (pure RGB).
        """
        return

    @abstractmethod
    def faces(self) -> list[Face]:
        """Get a list of all faces within the body.

        Returns
        -------
        list[Face]
        """
        return

    @abstractmethod
    def edges(self) -> list[Edge]:
        """Get a list of all edges within the body.

        Returns
        -------
        list[Edge]
        """
        return

    @abstractmethod
    def is_alive(self) -> bool:
        """Check if the body is still alive and has not been deleted."""
        return

    @abstractmethod
    def is_surface(self) -> bool:
        """Check if the body is a planar body."""
        return

    @abstractmethod
    def surface_thickness(self) -> Quantity | None:
        """Get the surface thickness of a surface body.

        Notes
        -----
        This method is only for surface-type bodies that have been assigned a surface thickness.
        """
        return

    @abstractmethod
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:
        """Get the surface offset type of a surface body.

        Notes
        -----
        This method is only for surface-type bodies that have been assigned a surface offset.
        """
        return

    @abstractmethod
    def volume(self) -> Quantity:
        """Calculate the volume of the body.

        Notes
        -----
        When dealing with a planar surface, a value of ``0`` is returned as a volume.
        """
        return

    @abstractmethod
    def material(self) -> Material:
        """Get the assigned material of the body.

        Returns
        -------
        Material
            Material assigned to the body.
        """
        return

    @abstractmethod
    def bounding_box(self) -> BoundingBox:
        """Get the bounding box of the body.

        Returns
        -------
        BoundingBox
            Bounding box of the body.
        """
        return

    @abstractmethod
    def assign_material(self, material: Material) -> None:
        """Assign a material against the active design.

        Parameters
        ----------
        material : Material
            Source material data.
        """
        return

    @abstractmethod
    def get_assigned_material(self) -> Material:
        """Get the assigned material of the body.

        Returns
        -------
        Material
            Material assigned to the body.
        """
        return

    @abstractmethod
    def add_midsurface_thickness(self, thickness: Quantity) -> None:
        """Add a mid-surface thickness to a surface body.

        Parameters
        ----------
        thickness : ~pint.Quantity
            Thickness to assign.

        Notes
        -----
        Only surface bodies are eligible for mid-surface thickness assignment.
        """
        return

    @abstractmethod
    def add_midsurface_offset(self, offset: "MidSurfaceOffsetType") -> None:
        """Add a mid-surface offset to a surface body.

        Parameters
        ----------
        offset_type : MidSurfaceOffsetType
            Surface offset to assign.

        Notes
        -----
        Only surface bodies are eligible for mid-surface offset assignment.
        """
        return

    @abstractmethod
    def imprint_curves(self, faces: list[Face], sketch: Sketch) -> tuple[list[Edge], list[Face]]:
        """Imprint all specified geometries onto specified faces of the body.

        Parameters
        ----------
        faces: list[Face]
            list of faces to imprint the curves of the sketch onto.
        sketch: Sketch
            All curves to imprint on the faces.

        Returns
        -------
        tuple[list[Edge], list[Face]]
            All impacted edges and faces from the imprint operation.
        """
        return

    @abstractmethod
    def project_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        """Project all specified geometries onto the body.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the projection.
        sketch: Sketch
            All curves to project on the body.
        closest_face: bool
            Whether to target the closest face with the projection.
        only_one_curve: bool, default: False
            Whether to project only one curve of the entire sketch. When
            ``True``, only one curve is projected.

        Returns
        -------
        list[Face]
            All faces from the project curves operation.

        Notes
        -----
        The ``only_one_curve`` parameter allows you to optimize the server call because
        projecting curves is an expensive operation. This reduces the workload on the
        server side.
        """
        return

    @abstractmethod
    def imprint_projected_curves(
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:  # noqa: D102
        """Project and imprint specified geometries onto the body.

        This method combines the ``project_curves()`` and ``imprint_curves()`` method into
        one method. It has higher performance than calling them back-to-back when dealing
        with many curves. Because it is a specialized function, this method only returns
        the faces (and not the edges) from the imprint operation.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the projection.
        sketch: Sketch
            All curves to project on the body.
        closest_face: bool
            Whether to target the closest face with the projection.
        only_one_curve: bool, default: False
            Whether to project only one curve of the entire sketch. When
            ``True``, only one curve is projected.

        Returns
        -------
        list[Face]
            All imprinted faces from the operation.

        Notes
        -----
        The ``only_one_curve`` parameter allows you to optimize the server call because
        projecting curves is an expensive operation. This reduces the workload on the
        server side.
        """
        return

    @abstractmethod
    def translate(self, direction: UnitVector3D, distance: Quantity | Distance | Real) -> None:
        """Translate the body in a specified direction and distance.

        Parameters
        ----------
        direction: UnitVector3D
            Direction of the translation.
        distance: ~pint.Quantity | Distance | Real
            Distance (magnitude) of the translation.

        Returns
        -------
        None
        """
        return

    @abstractmethod
    def rotate(
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Quantity | Angle | Real,
    ) -> None:
        """Rotate the geometry body around the specified axis by a given angle.

        Parameters
        ----------
        axis_origin: Point3D
            Origin of the rotational axis.
        axis_direction: UnitVector3D
            The axis of rotation.
        angle: ~pint.Quantity | Angle | Real
            Angle (magnitude) of the rotation.

        Returns
        -------
        None
        """
        return

    @abstractmethod
    def scale(self, value: Real) -> None:
        """Scale the geometry body by the given value.

        Parameters
        ----------
        value: Real
            Value to scale the body by.

        Notes
        -----
        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.
        """
        return

    @abstractmethod
    def map(self, frame: Frame) -> None:
        """Map the geometry body to the new specified frame.

        Parameters
        ----------
        frame: Frame
            Structure defining the orientation of the body.

        Notes
        -----
        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.
        """
        return

    @abstractmethod
    def mirror(self, plane: Plane) -> None:
        """Mirror the geometry body across the specified plane.

        Parameters
        ----------
        plane: Plane
            Represents the mirror.

        Notes
        -----
        The calling object is directly modified when this method is called.
        Thus, it is important to make copies if needed.
        """
        return

    @abstractmethod
    def get_collision(self, body: "Body") -> CollisionType:
        """Get the collision state between bodies.

        Parameters
        ----------
        body: Body
            Object that the collision state is checked with.

        Returns
        -------
        CollisionType
            Enum that defines the collision state between bodies.
        """
        return

    @abstractmethod
    def copy(self, parent: "Component", name: str = None) -> "Body":
        """Create a copy of the body under the specified parent.

        Parameters
        ----------
        parent: Component
            Parent component to place the new body under within the design assembly.
        name: str
            Name to give the new body.

        Returns
        -------
        Body
            Copy of the body.
        """
        return

    @abstractmethod
    def tessellate(
        self, merge: bool = False, tess_options: TessellationOptions | None = None
    ) -> Union["PolyData", "MultiBlock"]:
        """Tessellate the body and return the geometry as triangles.

        Parameters
        ----------
        merge : bool, default: False
            Whether to merge the body into a single mesh. When ``False`` (default), the
            number of triangles are preserved and only the topology is merged.
            When ``True``, the individual faces of the tessellation are merged.
        tess_options : TessellationOptions | None, default: None
            A set of options to determine the tessellation quality.

        Returns
        -------
        ~pyvista.PolyData, ~pyvista.MultiBlock
            Merged :class:`pyvista.PolyData` if ``merge=True`` or a composite dataset.

        Examples
        --------
        Extrude a box centered at the origin to create a rectangular body and
        tessellate it:

        >>> from ansys.geometry.core.misc.units import UNITS as u
        >>> from ansys.geometry.core.sketch import Sketch
        >>> from ansys.geometry.core.math import Plane, Point2D, Point3D, UnitVector3D
        >>> from ansys.geometry.core import Modeler
        >>> modeler = Modeler()
        >>> origin = Point3D([0, 0, 0])
        >>> plane = Plane(origin, direction_x=[1, 0, 0], direction_y=[0, 0, 1])
        >>> sketch = Sketch(plane)
        >>> box = sketch.box(Point2D([2, 0]), 4, 4)
        >>> design = modeler.create_design("my-design")
        >>> my_comp = design.add_component("my-comp")
        >>> body = my_comp.extrude_sketch("my-sketch", sketch, 1 * u.m)
        >>> blocks = body.tessellate()
        >>> blocks
        >>> MultiBlock(0x7F94EC757460)
             N Blocks:	6
             X Bounds:	0.000, 4.000
             Y Bounds:	-1.000, 0.000
             Z Bounds:	-0.500, 4.500

        Merge the body:

        >>> mesh = body.tessellate(merge=True)
        >>> mesh
        PolyData (0x7f94ec75f3a0)
          N Cells:	12
          N Points:	24
          X Bounds:	0.000e+00, 4.000e+00
          Y Bounds:	-1.000e+00, 0.000e+00
          Z Bounds:	-5.000e-01, 4.500e+00
          N Arrays:	0
        """
        return

    @abstractmethod
    def shell_body(self, offset: Real) -> bool:
        """Shell the body to the thickness specified.

        Parameters
        ----------
        offset : Real
            Shell thickness.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        return

    @abstractmethod
    def remove_faces(self, selection: Face | Iterable[Face], offset: Real) -> bool:
        """Shell by removing a given set of faces.

        Parameters
        ----------
        selection : Face | Iterable[Face]
            Face or faces to be removed.
        offset : Real
            Shell thickness.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        return

    @abstractmethod
    def plot(
        self,
        merge: bool = True,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        **plotting_options: dict | None,
    ) -> None:
        """Plot the body.

        Parameters
        ----------
        merge : bool, default: True
            Whether to merge the body into a single mesh. Performance improved when ``True``.
            When ``True`` (default), the individual faces of the tessellation are merged.
            When ``False``, the number of triangles are preserved and only the topology
            is merged.
        screenshot : str, default: None
            Path for saving a screenshot of the image that is being represented.
        use_trame : bool, default: None
            Whether to enable the use of `trame <https://kitware.github.io/trame/index.html>`_.
            The default is ``None``, in which case the
            ``ansys.tools.visualization_interface.USE_TRAME`` global setting is used.
        use_service_colors : bool, default: None
            Whether to use the colors assigned to the body in the service. The default
            is ``None``, in which case the ``ansys.geometry.core.USE_SERVICE_COLORS``
            global setting is used.
        **plotting_options : dict, default: None
            Keyword arguments for plotting. For allowable keyword arguments, see the
            :meth:`Plotter.add_mesh <pyvista.Plotter.add_mesh>` method.

        Examples
        --------
        Extrude a box centered at the origin to create rectangular body and
        plot it:

        >>> from ansys.geometry.core.misc.units import UNITS as u
        >>> from ansys.geometry.core.sketch import Sketch
        >>> from ansys.geometry.core.math import Plane, Point2D, Point3D, UnitVector3D
        >>> from ansys.geometry.core import Modeler
        >>> modeler = Modeler()
        >>> origin = Point3D([0, 0, 0])
        >>> plane = Plane(origin, direction_x=[1, 0, 0], direction_y=[0, 0, 1])
        >>> sketch = Sketch(plane)
        >>> box = sketch.box(Point2D([2, 0]), 4, 4)
        >>> design = modeler.create_design("my-design")
        >>> mycomp = design.add_component("my-comp")
        >>> body = mycomp.extrude_sketch("my-sketch", sketch, 1 * u.m)
        >>> body.plot()

        Plot the body and color each face individually:

        >>> body.plot(multi_colors=True)
        """
        return

    def intersect(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:
        """Intersect two (or more) bodies.

        Parameters
        ----------
        other : Body
            Body to intersect with.
        keep_other : bool, default: False
            Whether to retain the intersected body or not.

        Raises
        ------
        ValueError
            If the bodies do not intersect.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed. If the ``keep_other`` parameter is set to ``True``,
        the intersected body is retained.
        """
        return

    def subtract(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:
        """Subtract two (or more) bodies.

        Parameters
        ----------
        other : Body
            Body to subtract from the ``self`` parameter.
        keep_other : bool, default: False
            Whether to retain the subtracted body or not.

        Raises
        ------
        ValueError
            If the subtraction results in an empty (complete) subtraction.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed. If the ``keep_other`` parameter is set to ``True``,
        the subtracted body is retained.
        """
        return

    def unite(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:
        """Unite two (or more) bodies.

        Parameters
        ----------
        other : Body
            Body to unite with the ``self`` parameter.
        keep_other : bool, default: False
            Whether to retain the united body or not.

        Notes
        -----
        The ``self`` parameter is directly modified with the result, and
        the ``other`` parameter is consumed. Thus, it is important to make
        copies if needed. If the ``keep_other`` parameter is set to ``True``,
        the united body is retained.
        """
        return


class MasterBody(IBody):
    """Represents solids and surfaces organized within the design assembly.

    Solids and surfaces synchronize to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    parent_component : Component
        Parent component to place the new component under within the design assembly.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    is_surface : bool, default: False
        Whether the master body is a surface or an 3D object (with volume). The default
        is ``False``, in which case the master body is a surface. When ``True``, the
        master body is a 3D object (with volume).
    """

    def __init__(
        self,
        id: str,
        name: str,
        grpc_client: GrpcClient,
        is_surface: bool = False,
    ):
        """Initialize the ``MasterBody`` class."""
        check_type(id, str)
        check_type(name, str)
        check_type(grpc_client, GrpcClient)
        check_type(is_surface, bool)

        self._id = id
        self._name = name
        self._grpc_client = grpc_client
        self._is_surface = is_surface
        self._surface_thickness = None
        self._surface_offset = None
        self._is_alive = True
        self._commands_stub = CommandsStub(self._grpc_client.channel)
        self._tessellation = None
        self._fill_style = FillStyle.DEFAULT
        self._color = None

    def reset_tessellation_cache(func):  # noqa: N805
        """Decorate ``MasterBody`` methods that need tessellation cache update.

        Parameters
        ----------
        func : method
            Method to call.

        Returns
        -------
        Any
            Output of the method, if any.
        """

        @wraps(func)
        def wrapper(self: "MasterBody", *args, **kwargs):
            self._tessellation = None
            return func(self, *args, **kwargs)

        return wrapper

    @property
    def id(self) -> str:  # noqa: D102
        return self._id

    @cached_property
    def _grpc_id(self) -> EntityIdentifier:  # noqa: D102
        return EntityIdentifier(id=self._id)

    @property
    def name(self) -> str:  # noqa: D102
        return self._name

    @name.setter
    def name(self, value: str):  # noqa: D102
        self.set_name(value)

    @property
    def fill_style(self) -> str:  # noqa: D102
        return self._fill_style

    @fill_style.setter
    def fill_style(self, value: FillStyle):  # noqa: D102
        self.set_fill_style(value)

    @property
    @min_backend_version(25, 2, 0)
    def is_suppressed(self) -> bool:  # noqa: D102
        response = self._grpc_client.services.bodies.is_suppressed(id=self.id)
        return response.get("result")

    @is_suppressed.setter
    def is_suppressed(self, value: bool):  # noqa: D102
        self.set_suppressed(value)

    @property
    def color(self) -> str:  # noqa: D102
        """Get the current color of the body."""
        if self._color is None and self.is_alive:
            # Assigning default value first
            self._color = DEFAULT_COLOR

            if self._grpc_client.backend_version < (25, 1, 0):  # pragma: no cover
                # Server does not support color retrieval before version 25.1.0
                self._grpc_client.log.warning(
                    "Server does not support color retrieval. Default value assigned..."
                )
            else:
                # Fetch color from the server if it's not cached
                color = self._grpc_client.services.bodies.get_color(id=self.id).get("color")
                if color:
                    self._color = mcolors.to_hex(color, keep_alpha=True)

        return self._color

    @property
    def opacity(self) -> float:  # noqa: D102
        opacity_hex = self.color[7:]
        return int(opacity_hex, 16) / 255 if opacity_hex else 1

    @color.setter
    def color(self, value: str | tuple[float, float, float]) -> None:  # noqa: D102
        self.set_color(value)

    @opacity.setter
    def opacity(self, value: float) -> None:  # noqa: D102
        self.set_opacity(value)

    @property
    def is_surface(self) -> bool:  # noqa: D102
        return self._is_surface

    @property
    def surface_thickness(self) -> Quantity | None:  # noqa: D102
        return self._surface_thickness if self.is_surface else None

    @property
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._surface_offset if self.is_surface else None

    @property
    def faces(self) -> list[Face]:  # noqa: D102
        return self._get_faces_from_id(self)

    def _get_faces_from_id(self, body: Union["Body", "MasterBody"]) -> list[Face]:
        """Retrieve faces from a body ID."""
        self._grpc_client.log.debug(f"Retrieving faces for body {body.id} from server.")
        response = self._grpc_client.services.bodies.get_faces(id=body.id)
        return [
            Face(
                face_resp.get("id"),
                SurfaceType(face_resp.get("surface_type")),
                body,
                self._grpc_client,
                face_resp.get("is_reversed"),
            )
            for face_resp in response.get("faces")
        ]

    @property
    def edges(self) -> list[Edge]:  # noqa: D102
        return self._get_edges_from_id(self)

    def _get_edges_from_id(self, body: Union["Body", "MasterBody"]) -> list[Edge]:
        """Retrieve edges from a body ID."""
        self._grpc_client.log.debug(f"Retrieving edges for body {body.id} from server.")
        response = self._grpc_client.services.bodies.get_edges(id=body.id)
        return [
            Edge(
                edge_resp.get("id"),
                CurveType(edge_resp.get("curve_type")),
                body,
                self._grpc_client,
                edge_resp.get("is_reversed"),
            )
            for edge_resp in response.get("edges")
        ]

    @property
    def is_alive(self) -> bool:  # noqa: D102
        return self._is_alive

    @property
    def volume(self) -> Quantity:  # noqa: D102
        if self.is_surface:
            self._grpc_client.log.debug("Dealing with planar surface. Returning 0 as the volume.")
            return Quantity(0, DEFAULT_UNITS.SERVER_VOLUME)
        else:
            self._grpc_client.log.debug(f"Retrieving volume for body {self.id} from server.")
            response = self._grpc_client.services.bodies.get_volume(id=self.id)
            return response.get("volume")

    @property
    def material(self) -> Material:  # noqa: D102
        return self.get_assigned_material()

    @material.setter
    def material(self, value: Material):  # noqa: D102
        self.assign_material(value)

    @property
    @min_backend_version(25, 2, 0)
    def bounding_box(self) -> BoundingBox:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving bounding box for body {self.id} from server.")
        response = self._grpc_client.services.bodies.get_bounding_box(id=self.id)
        return BoundingBox(
            min_corner=response.get("min"),
            max_corner=response.get("max"),
            center=response.get("center"),
        )

    @check_input_types
    def assign_material(self, material: Material) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Assigning body {self.id} material {material.name}.")
        self._grpc_client.services.bodies.set_assigned_material(id=self.id, material=material)

    def get_assigned_material(self) -> Material:  # noqa: D102
        self._grpc_client.log.debug(f"Retrieving assigned material for body {self.id}.")
        response = self._grpc_client.services.bodies.get_assigned_material(id=self.id)
        return response.get("material")

    @protect_grpc
    @check_input_types
    def add_midsurface_thickness(self, thickness: Quantity) -> None:  # noqa: D102
        if self.is_surface:
            self._commands_stub.AssignMidSurfaceThickness(
                AssignMidSurfaceThicknessRequest(
                    bodies_or_faces=[self.id],
                    thickness=thickness.m_as(DEFAULT_UNITS.SERVER_LENGTH),
                )
            )
            self._surface_thickness = thickness
        else:
            self._grpc_client.log.warning(
                f"Body {self.name} cannot be assigned a mid-surface thickness because it is not a surface. Ignoring request."  # noqa : E501
            )

    @protect_grpc
    @check_input_types
    def add_midsurface_offset(self, offset: MidSurfaceOffsetType) -> None:  # noqa: D102
        if self.is_surface:
            self._commands_stub.AssignMidSurfaceOffsetType(
                AssignMidSurfaceOffsetTypeRequest(
                    bodies_or_faces=[self.id], offset_type=offset.value
                )
            )
            self._surface_offset = offset
        else:
            self._grpc_client.log.warning(
                f"Body {self.name} cannot be assigned a mid-surface offset because it is not a surface. Ignoring request."  # noqa : E501
            )

    @check_input_types
    def imprint_curves(  # noqa: D102
        self, faces: list[Face], sketch: Sketch
    ) -> tuple[list[Edge], list[Face]]:
        raise NotImplementedError(
            """
            imprint_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @check_input_types
    def project_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        raise NotImplementedError(
            """
            project_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @check_input_types
    def imprint_projected_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        raise NotImplementedError(
            """
            imprint_projected_curves is not implemented at the MasterBody level.
            Instead, call this method on a body.
            """
        )

    @check_input_types
    @reset_tessellation_cache
    def translate(  # noqa: D102
        self, direction: UnitVector3D, distance: Quantity | Distance | Real
    ) -> None:
        distance = distance if isinstance(distance, Distance) else Distance(distance)
        self._grpc_client.log.debug(f"Translating body {self.id}.")
        self._grpc_client.services.bodies.translate(
            ids=[self.id], direction=direction, distance=distance
        )

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_name(  # noqa: D102
        self, name: str
    ) -> None:
        self._grpc_client.log.debug(f"Renaming body {self.id} from '{self.name}' to '{name}'.")
        self._grpc_client.services.bodies.set_name(id=self.id, name=name)
        self._name = name

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_fill_style(  # noqa: D102
        self, fill_style: FillStyle
    ) -> None:
        self._grpc_client.log.debug(f"Setting body fill style {self.id}.")
        self._grpc_client.services.bodies.set_fill_style(id=self.id, fill_style=fill_style)
        self._fill_style = fill_style

    @check_input_types
    @min_backend_version(25, 2, 0)
    def set_suppressed(  # noqa: D102
        self, suppressed: bool
    ) -> None:
        """Set the body suppression state."""
        self._grpc_client.log.debug(f"Setting body {self.id}, as suppressed: {suppressed}.")
        self._grpc_client.services.bodies.set_suppressed(bodies=[self.id], is_suppressed=suppressed)

    @check_input_types
    @min_backend_version(25, 1, 0)
    def set_color(
        self, color: str | tuple[float, float, float] | tuple[float, float, float, float]
    ) -> None:
        """Set the color of the body."""
        self._grpc_client.log.debug(f"Setting body color of {self.id} to {color}.")
        color = convert_color_to_hex(color)
        self._grpc_client.services.bodies.set_color(id=self.id, color=color)
        self._color = color

    @check_input_types
    @min_backend_version(25, 2, 0)
    def set_opacity(self, opacity: float) -> None:
        """Set the opacity of the body."""
        self._grpc_client.log.debug(f"Setting body opacity of {self.id} to {opacity}.")
        opacity = convert_opacity_to_hex(opacity)
        new_color = self.color[0:7] + opacity
        self.set_color(new_color)

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def rotate(  # noqa: D102
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Quantity | Angle | Real,
    ) -> None:
        angle = angle if isinstance(angle, Angle) else Angle(angle)
        self._grpc_client.log.debug(f"Rotating body {self.id}.")
        self._grpc_client.services.bodies.rotate(
            id=self.id, axis_origin=axis_origin, axis_direction=axis_direction, angle=angle
        )

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def scale(self, value: Real) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Scaling body {self.id}.")
        self._grpc_client.services.bodies.scale(id=self.id, value=value)

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def map(self, frame: Frame) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Mapping body {self.id}.")
        self._grpc_client.services.bodies.map(id=self.id, frame=frame)

    @check_input_types
    @reset_tessellation_cache
    @min_backend_version(24, 2, 0)
    def mirror(self, plane: Plane) -> None:  # noqa: D102
        self._grpc_client.log.debug(f"Mirroring body {self.id}.")
        self._grpc_client.services.bodies.mirror(id=self.id, plane=plane)

    @min_backend_version(24, 2, 0)
    def get_collision(self, body: "Body") -> CollisionType:  # noqa: D102
        self._grpc_client.log.debug(f"Get collision between body {self.id} and body {body.id}.")
        response = self._grpc_client.services.bodies.get_collision(id=self.id, other_id=body.id)
        return CollisionType(response.get("collision_type"))

    def copy(self, parent: "Component", name: str = None) -> "Body":  # noqa: D102
        from ansys.geometry.core.designer.component import Component

        # Check input types
        check_type(parent, Component)
        copy_name = self.name if name is None else name
        check_type(copy_name, str)

        self._grpc_client.log.debug(f"Copying body {self.id}.")
        response = self._grpc_client.services.bodies.copy(
            id=self.id, parent_id=parent.id, name=copy_name
        )

        # Assign the new body to its specified parent (and return the new body)
        tb = MasterBody(
            response.get("master_id"), copy_name, self._grpc_client, is_surface=self.is_surface
        )
        parent._master_component.part.bodies.append(tb)
        parent._clear_cached_bodies()
        body_id = f"{parent.id}/{tb.id}" if parent.parent_component else tb.id
        return Body(body_id, response.get("name"), parent, tb)

    @graphics_required
    def tessellate(  # noqa: D102
        self,
        merge: bool = False,
        transform: Matrix44 = IDENTITY_MATRIX44,
        tess_options: TessellationOptions | None = None,
    ) -> Union["PolyData", "MultiBlock"]:
        # lazy import here to improve initial module load time
        import pyvista as pv

        if not self.is_alive:
            return pv.PolyData() if merge else pv.MultiBlock()

        # If the server does not support tessellation options, ignore them
        if tess_options is not None and self._grpc_client.backend_version < (25, 2, 0):
            self._grpc_client.log.warning(
                "Tessellation options are not supported by server"
                f" version {self._grpc_client.backend_version}. Ignoring options."
            )
            tess_options = None

        self._grpc_client.log.debug(f"Requesting tessellation for body {self.id}.")

        # cache tessellation
        if not self._tessellation:
            if tess_options is not None:
                response = self._grpc_client.services.bodies.get_tesellation_with_options(
                    id=self.id,
                    options=tess_options,
                )
            else:
                response = self._grpc_client.services.bodies.get_tesellation(
                    id=self.id,
                    backend_version=self._grpc_client.backend_version,
                )

            self._tessellation = response.get("tessellation")

        pdata = [tess.transform(transform, inplace=False) for tess in self._tessellation.values()]
        comp = pv.MultiBlock(pdata)

        if merge:
            ugrid = comp.combine()
            return pv.PolyData(var_inp=ugrid.points, faces=ugrid.cells)
        else:
            return comp

    @protect_grpc
    @reset_tessellation_cache
    @check_input_types
    @min_backend_version(25, 2, 0)
    def shell_body(self, offset: Real) -> bool:  # noqa: D102
        self._grpc_client.log.debug(f"Shelling body {self.id} to offset {offset}.")

        result = self._commands_stub.Shell(
            ShellRequest(
                selection=self._grpc_id,
                offset=offset,
            )
        )

        if result.success is False:
            self._grpc_client.log.warning(f"Failed to shell body {self.id}.")

        return result.success

    @protect_grpc
    @reset_tessellation_cache
    @check_input_types
    @min_backend_version(25, 2, 0)
    def remove_faces(self, selection: Face | Iterable[Face], offset: Real) -> bool:  # noqa: D102
        selection: list[Face] = selection if isinstance(selection, Iterable) else [selection]
        check_type_all_elements_in_iterable(selection, Face)

        # check if faces belong to this body
        for face in selection:
            if face.body.id != self.id:
                raise ValueError(f"Face {face.id} does not belong to body {self.id}.")

        self._grpc_client.log.debug(f"Removing faces to shell body {self.id}.")

        result = self._commands_stub.RemoveFaces(
            RemoveFacesRequest(
                selection=[face._grpc_id for face in selection],
                offset=offset,
            )
        )

        if result.success is False:
            self._grpc_client.log.warning(f"Failed to remove faces from body {self.id}.")

        return result.success

    def plot(  # noqa: D102
        self,
        merge: bool = True,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        **plotting_options: dict | None,
    ) -> None:
        raise NotImplementedError(
            "MasterBody does not implement plot methods. Call this method on a body instead."
        )

    def intersect(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def subtract(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def unite(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        raise NotImplementedError(
            "MasterBody does not implement Boolean methods. Call this method on a body instead."
        )

    def __repr__(self) -> str:
        """Represent the master body as a string."""
        lines = [f"ansys.geometry.core.designer.MasterBody {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Surface body         : {self.is_surface}")
        if self.is_surface:
            lines.append(f"  Surface thickness    : {self.surface_thickness}")
            lines.append(f"  Surface offset       : {self.surface_offset}")

        return "\n".join(lines)


class Body(IBody):
    """Represents solids and surfaces organized within the design assembly.

    Solids and surfaces synchronize to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    parent_component : Component
        Parent component to place the new component under within the design assembly.
    template : MasterBody
        Master body that this body is an occurrence of.
    """

    def __init__(self, id, name, parent_component: "Component", template: MasterBody) -> None:
        """Initialize the ``Body`` class."""
        self._id = id
        self._name = name
        self._parent_component = parent_component
        self._template = template

    def reset_tessellation_cache(func):  # noqa: N805
        """Decorate ``Body`` methods that require a tessellation cache update.

        Parameters
        ----------
        func : method
            Method to call.

        Returns
        -------
        Any
            Output of the method, if any.
        """

        @wraps(func)
        def wrapper(self: "Body", *args, **kwargs):
            self._reset_tessellation_cache()
            return func(self, *args, **kwargs)

        return wrapper

    def _reset_tessellation_cache(self):  # noqa: N805
        """Reset the cached tessellation for a body."""
        self._template._tessellation = None
        # if this reference is stale, reset the real cache in the part
        # this gets the matching id master body in the part
        master_in_part = next(
            (
                b
                for b in self.parent_component._master_component.part.bodies
                if b.id == self._template.id
            ),
            None,
        )
        if master_in_part is not None:
            master_in_part._tessellation = None

    @property
    def id(self) -> str:  # noqa: D102
        return self._id

    @cached_property
    def _grpc_id(self) -> EntityIdentifier:  # noqa: D102
        return EntityIdentifier(id=self._id)

    @property
    def name(self) -> str:  # noqa: D102
        return self._template.name

    @name.setter
    def name(self, value: str):  # noqa: D102
        self._template.name = value

    @property
    def fill_style(self) -> str:  # noqa: D102
        return self._template.fill_style

    @fill_style.setter
    def fill_style(self, fill_style: FillStyle) -> str:  # noqa: D102
        self._template.fill_style = fill_style

    @property
    def is_suppressed(self) -> bool:  # noqa: D102
        return self._template.is_suppressed

    @is_suppressed.setter
    def is_suppressed(self, suppressed: bool):  # noqa: D102
        self._template.is_suppressed = suppressed

    @property
    def color(self) -> str:  # noqa: D102
        return self._template.color

    @property
    def opacity(self) -> float:  # noqa: D102
        return self._template.opacity

    @color.setter
    def color(self, color: str | tuple[float, float, float]) -> None:  # noqa: D102
        return self._template.set_color(color)

    @opacity.setter
    def opacity(self, opacity: float) -> None:  # noqa: D102
        return self._template.set_opacity(opacity)

    @property
    def parent_component(self) -> "Component":  # noqa: D102
        return self._parent_component

    @property
    @ensure_design_is_active
    def faces(self) -> list[Face]:  # noqa: D102
        return self._template._get_faces_from_id(self)

    @property
    @ensure_design_is_active
    def edges(self) -> list[Edge]:  # noqa: D102
        return self._template._get_edges_from_id(self)

    @property
    def _is_alive(self) -> bool:  # noqa: D102
        return self._template.is_alive

    @_is_alive.setter
    def _is_alive(self, value: bool):  # noqa: D102
        self._template._is_alive = value

    @property
    def is_alive(self) -> bool:  # noqa: D102
        return self._is_alive

    @property
    def is_surface(self) -> bool:  # noqa: D102
        return self._template.is_surface

    @property
    def _surface_thickness(self) -> Quantity | None:  # noqa: D102
        return self._template.surface_thickness

    @_surface_thickness.setter
    def _surface_thickness(self, value):  # noqa: D102
        self._template._surface_thickness = value

    @property
    def surface_thickness(self) -> Quantity | None:  # noqa: D102
        return self._surface_thickness

    @property
    def _surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._template._surface_offset

    @_surface_offset.setter
    def _surface_offset(self, value: "MidSurfaceOffsetType"):  # noqa: D102
        self._template._surface_offset = value

    @property
    def surface_offset(self) -> Union["MidSurfaceOffsetType", None]:  # noqa: D102
        return self._surface_offset

    @property
    @ensure_design_is_active
    def volume(self) -> Quantity:  # noqa: D102
        return self._template.volume

    @property
    @ensure_design_is_active
    def material(self) -> Material:  # noqa: D102
        return self._template.material

    @material.setter
    def material(self, value: Material):  # noqa: D102
        self._template.material = value

    @property
    def bounding_box(self) -> BoundingBox:  # noqa: D102
        return self._template.bounding_box

    @ensure_design_is_active
    def assign_material(self, material: Material) -> None:  # noqa: D102
        self._template.assign_material(material)

    @ensure_design_is_active
    def get_assigned_material(self) -> Material:  # noqa: D102
        return self._template.get_assigned_material()

    @ensure_design_is_active
    def add_midsurface_thickness(self, thickness: Quantity) -> None:  # noqa: D102
        self._template.add_midsurface_thickness(thickness)

    @ensure_design_is_active
    def add_midsurface_offset(  # noqa: D102
        self, offset: "MidSurfaceOffsetType"
    ) -> None:
        self._template.add_midsurface_offset(offset)

    @protect_grpc
    @ensure_design_is_active
    def imprint_curves(  # noqa: D102
        self, faces: list[Face], sketch: Sketch = None, trimmed_curves: list[TrimmedCurve] = None
    ) -> tuple[list[Edge], list[Face]]:
        """Imprint curves onto the specified faces using a sketch or edges.

        Parameters
        ----------
        faces : list[Face]
            The list of faces to imprint the curves onto.
        sketch : Sketch, optional
            The sketch containing curves to imprint.
        trimmed_curves : list[TrimmedCurve], optional
            The list of curves to be imprinted. If sketch is provided, this parameter is ignored.

        Returns
        -------
        tuple[list[Edge], list[Face]]
            A tuple containing the list of new edges and faces created by the imprint operation.
        """
        if sketch is None and self._template._grpc_client.backend_version < (25, 2, 0):
            raise ValueError(
                "A sketch must be provided for imprinting when using API versions below 25.2.0."
            )

        if sketch is None and trimmed_curves is None:
            raise ValueError("Either a sketch or edges must be provided for imprinting.")

        # Verify that each of the faces provided are part of this body
        body_faces = self.faces
        for provided_face in faces:
            if not any(provided_face.id == body_face.id for body_face in body_faces):
                raise ValueError(f"Face with ID {provided_face.id} is not part of this body.")

        self._template._grpc_client.log.debug(
            f"Imprinting curves on {self.id} for faces {[face.id for face in faces]}."
        )

        curves = None
        grpc_trimmed_curves = None

        if sketch:
            curves = sketch_shapes_to_grpc_geometries(sketch._plane, sketch.edges, sketch.faces)

        if trimmed_curves:
            grpc_trimmed_curves = [
                trimmed_curve_to_grpc_trimmed_curve(curve) for curve in trimmed_curves
            ]

        imprint_response = self._template._commands_stub.ImprintCurves(
            ImprintCurvesRequest(
                body=self._id,
                curves=curves,
                faces=[face._id for face in faces],
                plane=plane_to_grpc_plane(sketch.plane) if sketch else None,
                trimmed_curves=grpc_trimmed_curves,
            )
        )

        new_edges = [
            Edge(
                grpc_edge.id,
                CurveType(grpc_edge.curve_type),
                self,
                self._template._grpc_client,
            )
            for grpc_edge in imprint_response.edges
        ]

        new_faces = [
            Face(
                grpc_face.id,
                SurfaceType(grpc_face.surface_type),
                self,
                self._template._grpc_client,
            )
            for grpc_face in imprint_response.faces
        ]

        return (new_edges, new_faces)

    @protect_grpc
    @ensure_design_is_active
    def project_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        curves = sketch_shapes_to_grpc_geometries(
            sketch._plane, sketch.edges, sketch.faces, only_one_curve=only_one_curve
        )
        self._template._grpc_client.log.debug(f"Projecting provided curves on {self.id}.")

        project_response = self._template._commands_stub.ProjectCurves(
            ProjectCurvesRequest(
                body=self._id,
                curves=curves,
                direction=unit_vector_to_grpc_direction(direction),
                closest_face=closest_face,
                plane=plane_to_grpc_plane(sketch.plane),
            )
        )

        projected_faces = [
            Face(
                grpc_face.id,
                SurfaceType(grpc_face.surface_type),
                self,
                self._template._grpc_client,
            )
            for grpc_face in project_response.faces
        ]

        return projected_faces

    @check_input_types
    @protect_grpc
    @ensure_design_is_active
    def imprint_projected_curves(  # noqa: D102
        self,
        direction: UnitVector3D,
        sketch: Sketch,
        closest_face: bool,
        only_one_curve: bool = False,
    ) -> list[Face]:
        curves = sketch_shapes_to_grpc_geometries(
            sketch._plane, sketch.edges, sketch.faces, only_one_curve=only_one_curve
        )
        self._template._grpc_client.log.debug(f"Projecting provided curves on {self.id}.")

        response = self._template._commands_stub.ImprintProjectedCurves(
            ProjectCurvesRequest(
                body=self._id,
                curves=curves,
                direction=unit_vector_to_grpc_direction(direction),
                closest_face=closest_face,
                plane=plane_to_grpc_plane(sketch.plane),
            )
        )

        imprinted_faces = [
            Face(
                grpc_face.id,
                SurfaceType(grpc_face.surface_type),
                self,
                self._template._grpc_client,
            )
            for grpc_face in response.faces
        ]

        return imprinted_faces

    @ensure_design_is_active
    def set_name(self, name: str) -> None:  # noqa: D102
        return self._template.set_name(name)

    @ensure_design_is_active
    def set_fill_style(self, fill_style: FillStyle) -> None:  # noqa: D102
        return self._template.set_fill_style(fill_style)

    @ensure_design_is_active
    def set_suppressed(self, suppressed: bool) -> None:  # noqa: D102
        return self._template.set_suppressed(suppressed)

    @ensure_design_is_active
    def set_color(self, color: str | tuple[float, float, float]) -> None:  # noqa: D102
        return self._template.set_color(color)

    @ensure_design_is_active
    def translate(  # noqa: D102
        self, direction: UnitVector3D, distance: Quantity | Distance | Real
    ) -> None:
        return self._template.translate(direction, distance)

    @ensure_design_is_active
    def rotate(  # noqa: D102
        self,
        axis_origin: Point3D,
        axis_direction: UnitVector3D,
        angle: Quantity | Angle | Real,
    ) -> None:
        return self._template.rotate(axis_origin, axis_direction, angle)

    @ensure_design_is_active
    def scale(self, value: Real) -> None:  # noqa: D102
        return self._template.scale(value)

    @ensure_design_is_active
    def map(self, frame: Frame) -> None:  # noqa: D102
        return self._template.map(frame)

    @ensure_design_is_active
    def mirror(self, plane: Plane) -> None:  # noqa: D102
        return self._template.mirror(plane)

    @ensure_design_is_active
    def get_collision(self, body: "Body") -> CollisionType:  # noqa: D102
        return self._template.get_collision(body)

    @ensure_design_is_active
    def copy(self, parent: "Component", name: str = None) -> "Body":  # noqa: D102
        return self._template.copy(parent, name)

    @ensure_design_is_active
    def tessellate(  # noqa: D102
        self, merge: bool = False, tess_options: TessellationOptions | None = None
    ) -> Union["PolyData", "MultiBlock"]:
        return self._template.tessellate(
            merge, self.parent_component.get_world_transform(), tess_options
        )

    @ensure_design_is_active
    def shell_body(self, offset: Real) -> bool:  # noqa: D102
        return self._template.shell_body(offset)

    @ensure_design_is_active
    def remove_faces(self, selection: Face | Iterable[Face], offset: Real) -> bool:  # noqa: D102
        return self._template.remove_faces(selection, offset)

    @graphics_required
    def plot(  # noqa: D102
        self,
        merge: bool = True,
        screenshot: str | None = None,
        use_trame: bool | None = None,
        use_service_colors: bool | None = None,
        **plotting_options: dict | None,
    ) -> None:
        # lazy import here to improve initial module load time
        import ansys.geometry.core as pyansys_geometry
        from ansys.geometry.core.plotting import GeometryPlotter
        from ansys.tools.visualization_interface.types.mesh_object_plot import (
            MeshObjectPlot,
        )

        use_service_colors = (
            use_service_colors
            if use_service_colors is not None
            else pyansys_geometry.USE_SERVICE_COLORS
        )

        # Add to plotting options as well... to be used by the plotter if necessary
        plotting_options["merge_bodies"] = merge

        # In case that service colors are requested, merge will be ignored.
        # An info message will be issued to the user.
        if use_service_colors and merge:
            self._template._grpc_client.log.info(
                "Ignoring 'merge' option since 'use_service_colors' is set to True."
            )
            plotting_options["merge_bodies"] = False

        mesh_object = (
            self if use_service_colors else MeshObjectPlot(self, self.tessellate(merge=merge))
        )
        pl = GeometryPlotter(use_trame=use_trame, use_service_colors=use_service_colors)
        pl.plot(mesh_object, **plotting_options)
        pl.show(screenshot=screenshot, **plotting_options)

    def intersect(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        if self._template._grpc_client.backend_version < __TEMPORARY_BOOL_OPS_FIX__:
            self.__generic_boolean_op(other, keep_other, "intersect", "bodies do not intersect")
        else:
            self.__generic_boolean_command(
                other, keep_other, "intersect", "bodies do not intersect"
            )

    def subtract(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        if self._template._grpc_client.backend_version < __TEMPORARY_BOOL_OPS_FIX__:
            self.__generic_boolean_op(other, keep_other, "subtract", "empty (complete) subtraction")
        else:
            self.__generic_boolean_command(
                other, keep_other, "subtract", "empty (complete) subtraction"
            )

    def unite(self, other: Union["Body", Iterable["Body"]], keep_other: bool = False) -> None:  # noqa: D102
        if self._template._grpc_client.backend_version < __TEMPORARY_BOOL_OPS_FIX__:
            self.__generic_boolean_op(other, keep_other, "unite", "union operation failed")
        else:
            self.__generic_boolean_command(other, False, "unite", "union operation failed")

    @protect_grpc
    @reset_tessellation_cache
    @ensure_design_is_active
    @check_input_types
    def __generic_boolean_command(
        self,
        other: Union["Body", Iterable["Body"]],
        keep_other: bool,
        type_bool_op: str,
        err_bool_op: str,
    ) -> None:
        parent_design = get_design_from_body(self)
        other_bodies = other if isinstance(other, Iterable) else [other]
        if type_bool_op == "intersect":
            body_ids = [body._grpc_id for body in other_bodies]
            target_ids = [self._grpc_id]
            request = CombineIntersectBodiesRequest(
                target_selection=target_ids,
                tool_selection=body_ids,
                subtract_from_target=False,
                keep_cutter=keep_other,
            )
            response = self._template._commands_stub.CombineIntersectBodies(request)
        elif type_bool_op == "subtract":
            body_ids = [body._grpc_id for body in other_bodies]
            target_ids = [self._grpc_id]
            request = CombineIntersectBodiesRequest(
                target_selection=target_ids,
                tool_selection=body_ids,
                subtract_from_target=True,
                keep_cutter=keep_other,
            )
            response = self._template._commands_stub.CombineIntersectBodies(request)
        elif type_bool_op == "unite":
            bodies = [self]
            bodies.extend(other_bodies)
            body_ids = [body._grpc_id for body in bodies]
            request = CombineMergeBodiesRequest(target_selection=body_ids)
            response = self._template._commands_stub.CombineMergeBodies(request)
        else:
            raise ValueError("Unknown operation requested")
        if not response.success:
            raise ValueError(
                f"Operation of type '{type_bool_op}' failed: {err_bool_op}.\n"
                f"Involving bodies:{self}, {other_bodies}"
            )

        if not keep_other:
            for b in other_bodies:
                b.parent_component.delete_body(b)

        parent_design._update_design_inplace()

    @reset_tessellation_cache
    @ensure_design_is_active
    @check_input_types
    def __generic_boolean_op(
        self,
        other: Union["Body", Iterable["Body"]],
        keep_other: bool,
        method: str,
        err_msg: str,
    ) -> None:
        grpc_other = other if isinstance(other, Iterable) else [other]
        if keep_other:
            # Make a copy of the other body to keep it...
            # stored temporarily in the parent component - since it will be deleted
            grpc_other = [b.copy(self.parent_component, f"BoolOpCopy_{b.name}") for b in grpc_other]

        self._template._grpc_client.services.bodies.boolean(
            target=self,
            other=grpc_other,
            method=method,
            err_msg=err_msg,
        )

        for b in grpc_other:
            b.parent_component.delete_body(b)

    def __repr__(self) -> str:
        """Represent the ``Body`` as a string."""
        lines = [f"ansys.geometry.core.designer.Body {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(f"  Parent component     : {self._parent_component.name}")
        lines.append(f"  MasterBody           : {self._template.id}")
        lines.append(f"  Surface body         : {self.is_surface}")
        if self.is_surface:
            lines.append(f"  Surface thickness    : {self.surface_thickness}")
            lines.append(f"  Surface offset       : {self.surface_offset}")
        lines.append(f"  Color                : {self.color}")

        nl = "\n"
        return f"{nl}{nl.join(lines)}{nl}"



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the repair tools service implementation for v0.

This module defines an abstract base class for a gRPC-based repair tools service.
The class provides a set of abstract methods for identifying and repairing various
geometry issues, such as split edges, extra edges, duplicate faces etc.
"""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.repair_tools import GRPCRepairToolsService


class GRPCRepairToolsServiceV0(GRPCRepairToolsService):  # noqa: D102
    """Repair tools service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    Repair Tools service. It is specifically designed for the v0 version
    of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):
        from ansys.api.geometry.v0.repairtools_pb2_grpc import RepairToolsStub

        self.stub = RepairToolsStub(channel)

    @protect_grpc
    def find_split_edges(self, **kwargs) -> dict:  # noqa: D102
        from google.protobuf.wrappers_pb2 import DoubleValue

        from ansys.api.geometry.v0.repairtools_pb2 import FindSplitEdgesRequest

        # Create the gRPC request
        request = FindSplitEdgesRequest(
            bodies_or_faces=kwargs["bodies_or_faces"],
            angle=DoubleValue(value=float(kwargs["angle"])),
            distance=DoubleValue(value=float(kwargs["distance"])),
        )

        # Call the gRPC service
        response = self.stub.FindSplitEdges(request)

        # Format and return the response as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_extra_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindExtraEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindExtraEdgesRequest(selection=kwargs["selection"])

        # Return the response - formatted as a dictionary
        response = self.stub.FindExtraEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_inexact_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindInexactEdgesRequest

        request = FindInexactEdgesRequest(selection=kwargs["selection"])

        # Call the gRPC service
        response = self.stub.FindInexactEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_short_edges(self, **kwargs) -> dict:  # noqa: D102
        from google.protobuf.wrappers_pb2 import DoubleValue

        from ansys.api.geometry.v0.repairtools_pb2 import FindShortEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindShortEdgesRequest(
            selection=kwargs["selection"],
            max_edge_length=DoubleValue(value=kwargs["length"]),
        )

        # Call the gRPC service
        response = self.stub.FindShortEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_duplicate_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindDuplicateFacesRequest

        request = FindDuplicateFacesRequest(faces=kwargs["faces"])

        # Call the gRPC service
        response = self.stub.FindDuplicateFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "faces": res.face_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_missing_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindMissingFacesRequest

        request = FindMissingFacesRequest(faces=kwargs["faces"])
        # Call the gRPC service
        response = self.stub.FindMissingFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_small_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindSmallFacesRequest

        request = FindSmallFacesRequest(selection=kwargs["selection"])
        # Call the gRPC service
        response = self.stub.FindSmallFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "faces": res.face_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_stitch_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindStitchFacesRequest

        request = FindStitchFacesRequest(faces=kwargs["faces"])
        # Call the gRPC service
        response = self.stub.FindStitchFaces(request)
        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "bodies": res.body_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_simplify(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindAdjustSimplifyRequest

        request = FindAdjustSimplifyRequest(selection=kwargs["selection"])

        # Call the gRPC service
        response = self.stub.FindAdjustSimplify(request)
        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "bodies": res.body_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_and_fix_simplify(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindAdjustSimplifyRequest

        request = FindAdjustSimplifyRequest(
            selection=kwargs["selection"],
            comprehensive=kwargs["comprehensive_result"],
        )
        # Call the gRPC service
        response = self.stub.FindAndSimplify(request)
        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": [],
            "modified_bodies_monikers": [],
        }

    @protect_grpc
    def inspect_geometry(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import InspectGeometryRequest

        # Create the gRPC request
        request = InspectGeometryRequest(bodies=kwargs.get("bodies", []))

        # Call the gRPC service
        inspect_result_response = self.stub.InspectGeometry(request)

        # Serialize and return the response
        return self.serialize_inspect_result_response(inspect_result_response)

    @protect_grpc
    def repair_geometry(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import RepairGeometryRequest

        request = RepairGeometryRequest(bodies=kwargs.get("bodies", []))

        # Call the gRPC service
        response = self.stub.RepairGeometry(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.result.success,
        }

    @protect_grpc
    def find_interferences(self, **kwargs) -> dict:  # noqa: D102
        from google.protobuf.wrappers_pb2 import BoolValue

        from ansys.api.geometry.v0.repairtools_pb2 import FindInterferenceRequest

        request = FindInterferenceRequest(
            bodies=kwargs["bodies"],
            cut_smaller_body=BoolValue(value=kwargs["cut_smaller_body"]),
        )

        # Call the gRPC service
        response = self.stub.FindInterference(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "bodies": res.body_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_and_fix_short_edges(self, **kwargs):  # noqa: D102
        from google.protobuf.wrappers_pb2 import DoubleValue

        from ansys.api.geometry.v0.repairtools_pb2 import FindShortEdgesRequest

        request = FindShortEdgesRequest(
            selection=kwargs["selection"],
            max_edge_length=DoubleValue(value=kwargs["length"]),
            comprehensive=kwargs["comprehensive_result"],
        )
        # Call the gRPC service
        response = self.stub.FindAndFixShortEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": [],
            "modified_bodies_monikers": [],
        }

    @protect_grpc
    def find_and_fix_extra_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindExtraEdgesRequest

        request = FindExtraEdgesRequest(
            selection=kwargs["selection"],
            comprehensive=kwargs["comprehensive_result"],
        )
        # Call the gRPC service
        response = self.stub.FindAndFixExtraEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": response.created_bodies_monikers,
            "modified_bodies_monikers": response.modified_bodies_monikers,
        }

    @protect_grpc
    def find_and_fix_split_edges(self, **kwargs) -> dict:  # noqa: D102
        from google.protobuf.wrappers_pb2 import DoubleValue

        from ansys.api.geometry.v0.repairtools_pb2 import FindSplitEdgesRequest

        request = FindSplitEdgesRequest(
            bodies_or_faces=kwargs["bodies_or_faces"],
            angle=DoubleValue(value=float(kwargs["angle"])),
            distance=DoubleValue(value=float(kwargs["length"])),
            comprehensive=kwargs["comprehensive_result"],
        )

        # Call the gRPC service
        response = self.stub.FindAndFixSplitEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": [],
            "modified_bodies_monikers": [],
        }

    @staticmethod
    def serialize_inspect_result_response(response) -> dict:  # noqa: D102
        def serialize_body(body):
            return {
                "id": body.id,
                "name": body.name,
                "can_suppress": body.can_suppress,
                "transform_to_master": {
                    "m00": body.transform_to_master.m00,
                    "m11": body.transform_to_master.m11,
                    "m22": body.transform_to_master.m22,
                    "m33": body.transform_to_master.m33,
                },
                "master_id": body.master_id,
                "parent_id": body.parent_id,
            }

        def serialize_face(face):
            return {
                "id": face.id,
                "surface_type": face.surface_type,
                "export_id": face.export_id,
                "is_reversed": getattr(face, "is_reversed", False),
                "parent_id": face.parent_id.id,
            }

        def serialize_edge(edge):
            return {
                "id": edge.id,
                "curve_type": edge.curve_type,
                "export_id": edge.export_id,
                "length": edge.length,
                "owner_id": edge.owner_id,
                "parent": serialize_body(edge.parent) if hasattr(edge, "parent") else None,
            }

        def serialize_issue(issue):
            return {
                "message_type": issue.message_type,
                "message_id": issue.message_id,
                "message": issue.message,
                "faces": [serialize_face(f) for f in issue.faces],
                "edges": [serialize_edge(e) for e in issue.edges],
            }

        return {
            "issues_by_body": [
                {
                    "body": serialize_body(body_issues.body),
                    "issues": [serialize_issue(i) for i in body_issues.issues],
                }
                for body_issues in response.issues_by_body
            ]
        }
